/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file arg.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// You may want to include "textfile.h" and "timeused.h" before this file.
// Started on March 3, 2002. Last revision: March 9, 2007.


#ifndef _CAPD_ARG_H_
#define _CAPD_ARG_H_

#include "homology/config.h"

#include <iostream>
#include <iomanip>
#include <cstring>
#include <sstream>
#include <cctype>

namespace capd {
namespace homology {

// the main front-end class defined in this file
class arguments;


// --------------------------------------------------
// ------------------- argelement -------------------
// --------------------------------------------------

class argflags
{
public:
	enum names {obligatory = 0x01, hasdefault = 0x02,
		filled = 0x04, breakinterpreting = 0x08,
		toomany = 0x10, missingvalue = 0x20,
		readerror = 0x40, ignorevalue = 0x80};
	argflags (): n (0) {}
	~argflags () {}
	void set (int flag) {n |= flag;}
	void unset (int flag) {n &= ~flag;}
	int get (int flag) const {return !!(n & flag);}

protected:
	int n;
}; /* argflags */

class argelement
{
public:
	// default constructor and destructor
	argelement (const char *_name);

	// default destructor
	virtual ~argelement ();

	// return argument's name
	const char *getname () const;

	// get the argument's value string from the argument string
	// or return NULL if this is not this argument
	char *getvalue (char *str);

	// set the value of this argument according to the string;
	// if necessary, the next argument may be used;
	// return: 0 = Ok, -1 = Error, 1 = next used
	virtual int setvalue (char *str, char *next) = 0;

	// restore the previous argument value (except for tables)
	virtual void restore () = 0;

	// output the argument element to the output stream
	virtual void show (std::ostream &out) const = 0;

	// reset flags
	void resetflags ();

	// set, unset and get a specific flag
	void set (int flag);
	void unset (int flag);
	int get (int flag) const;

private:

	// the argument name (without '-'): dynamically allocated!
	char *name;

	// flags
	argflags flags;

}; /* class argelement */

inline argelement::argelement (const char *_name): flags ()
{
	if (!_name || !*_name)
		name = NULL;
	else
	{
		name = new char [strlen (_name) + 1];
		strcpy (name, _name);
	}
	return;
} /* argelement::argelement */

inline argelement::~argelement ()
{
	if (name)
		delete [] name;
	return;
} /* argelement::~argelement */

inline const char *argelement::getname () const
{
	return name;
} /* argelement::getname */

inline void argelement::resetflags ()
{
	flags. unset (argflags::filled);
	flags. unset (argflags::readerror);
	flags. unset (argflags::toomany);
	flags. unset (argflags::missingvalue);
	return;
} /* argelement::resetflags */

inline void argelement::set (int flag)
{
	flags. set (flag);
	return;
} /* argelement::set */

inline void argelement::unset (int flag)
{
	flags. unset (flag);
	return;
} /* argelement::unset */

inline int argelement::get (int flag) const
{
	return flags. get (flag);
} /* argelement::get */

inline std::ostream &operator << (std::ostream &out, const argelement &p)
{
	p. show (out);
	return out;
} /* operator << */


// --------------------------------------------------
// -------------------- argunit ---------------------
// --------------------------------------------------

template <class type>
class argunit: public argelement
{
public:
	// the only constructors
	argunit (const char *_name, type &_value);
	argunit (const char *_name, type &_value, type defaultvalue);
	argunit (const char *_name, type *_value, int &_count,
		int _size);
	argunit (const char *_name, type *_value, int &_count,
		int _size, type defaultvalue);

	// set the argument value from the text string
	int setvalue (char *str, char *next);

	// restore the previous argument value (except for tables)
	void restore ();

	// display the value and some information
	void show (std::ostream &out) const;

protected:
	type &value;
	type defaultvalue;
	type previousvalue;

	type *table;
	int *count, previouscount, size;

}; /* class argunit */

template <class type>
inline argunit<type>::argunit (const char *_name, type &_value):
	argelement (_name), value (_value), previousvalue (_value),
	table (NULL)
{
	return;
} /* argunit<type>::argunit */

template <class type>
inline argunit<type>::argunit (const char *_name, type &_value,
	type _defaultvalue):
	argelement (_name), value (_value), defaultvalue (_defaultvalue),
	previousvalue (_value), table (NULL)
{
	set (argflags::hasdefault);
	return;
} /* argunit<type>::argunit */

template <class type>
inline argunit<type>::argunit (const char *_name, type *_value,
	int &_count, int _size):
	argelement (_name), value (*_value), previousvalue (*_value),
	table (_value), count (&_count), previouscount (_count),
	size (_size)
{
	return;
} /* argunit<type>::argunit */

template <class type>
inline argunit<type>::argunit (const char *_name, type *_value,
	int &_count, int _size, type _defaultvalue):
	argelement (_name), value (*_value), defaultvalue (_defaultvalue),
	previousvalue (*_value), table (_value),
	count (&_count), previouscount (_count), size (_size)
	
{
	set (argflags::hasdefault);
	return;
} /* argunit<type>::argunit */

template <class type>
inline int readfromstring (char *str, type &t)
{
	std::istringstream s (str);
	try
	{
		s >> t;
		if (!s)
			return -1;
	}
	catch (...)
	{
		return -1;
	}
	return 0;
} /* readfromstring */

inline int readfromstring (char *str, char *&t)
{
	t = str;
	return 0;
} /* readfromstring */

inline int readfromstring (char *str, bool &t)
{
	switch (*str)
	{
		case 'T':
		case 't':
		case 'Y':
		case 'y':
		case '1':
			t = true;
			return 0;
		case 'F':
		case 'f':
		case 'N':
		case 'n':
		case '0':
			t = false;
			return 0;
		default:
			return -1;
	}
} /* readfromstring */

template <class type>
inline int argunit<type>::setvalue (char *str, char *next)
{
	// determine the string from which the value should be read
	int usenext = 0;
	char *s;
	if (str && *str)
		s = str;
	else if (next && *next && ((*next != '-') || std::isdigit (next [1])))
	{
		s = next;
		usenext = 1;
	}
	else
		s = NULL;

	// read the element from the string if available
	int result = -1;
	type element;
	if (s)
		result = readfromstring (s, element);

	// if could not read the string, try the default value
	if (result < 0)
		if (get (argflags::hasdefault))
		{
			element = defaultvalue;
			usenext = 0;
		}
		else
		{
			if (s)
				set (argflags::readerror);
			else
				set (argflags::missingvalue);
			return 0;
		}

	// put the element to its place
	if (table)
		// if there is still room in the table, put the element there
		if (*count < size)
			table [(*count) ++] = element;
		// if the table is full, indicate it
		else
		{
			set (argflags::toomany);
			set (argflags::filled);
		}
	else
	{
		value = element;
		set (argflags::filled);
	}

	return usenext;
} /* argunit<type>::setvalue */

template <class type>
void argunit<type>::restore ()
{
	if (!table)
		value = previousvalue;
	else
		*count = previouscount;
	resetflags ();
	return;
} /* argunit<type>::restore */

template <class type>
void argunit<type>::show (std::ostream &out) const
{
	if (get (argflags::obligatory))
		out << "Oblig. ";
	if (getname () && *getname ())
	{
		if (get (argflags::ignorevalue))
			out << "Switch";
		else
			out << "Param.";
		out << " '-" << getname () << "'";
	}
	else
		out << "Word";
	if (!table && get (argflags::filled))
		out << " set to '" << value << "'";
	else if (table && *count)
	{
		for (int i = 0; i < *count; i ++)
			out << (i ? ", " : " set to '") << table [i];
		out << "'";
		if (get (argflags::toomany))
			out << " [too many!]";
	}
	else
		out << " not found";
	if (get (argflags::hasdefault))
		out << " (default: '" << defaultvalue << "')";
	else
		out << " (no default value)";
	out << "." << std::endl;
	return;
} /* argunit<type>::show */


// --------------------------------------------------
// ------------------- arguments --------------------
// --------------------------------------------------
// This is the main class which gathers the command-line arguments
// and decodes them. It is recommended that you use the functions "arg..."
// defined below to enqueue the arguments into the list of arguments.
// When the list is complete, call the "analyze" function.
// See "argtest.cpp" and the documentation in "arg.txt" for details.

class arguments
{
public:
	// default constructor and destructor
	arguments ();
	~arguments ();

	// add an argument to the end of the list
	void add (argelement *p);

	// remove all the arguments with this name from the list
	// and return the number of removed items
	int remove (char *name);

	// analyze the arguments from the argument strings;
	// on error return negative code (minus number of errors)
	// and display error messages to the given output stream;
	// in case of break request return 1
	int analyze (int argc, char *argv [],
		std::ostream &out = std::cerr);

	// reset flags and return previous argument values
	// (except for tables)
	void reset ();

	// display the arguments actually read (for debugging)
	friend std::ostream &operator << (std::ostream &out,
		arguments &p);

protected:
	// the number of arguments and the length of the table
	int n, length;

	// table of pointers to specific arguments
	argelement **tab;

	// increase the tables if necessary
	void inctable ();

}; /* class arguments */

inline arguments::arguments ()
{
	n = 0;
	length = 0;
	tab = NULL;
	return;
} /* arguments::arguments */

inline void arguments::add (argelement *p)
{
	inctable ();
	tab [n ++] = p;
	return;
} /* arguments::add */

// --------------------------------------------------

template <class type>
inline void arg (arguments &a, const char *name, type &value)
{
	argunit<type> *p = new argunit<type> (name, value);
	a. add (p);
	return;
} /* arg */

template <class type>
inline void arg (arguments &a, const char *name, type &value,
	type defaultvalue)
{
	argunit<type> *p = new argunit<type> (name, value, defaultvalue);
	a. add (p);
	return;
} /* arg */

inline void arg (arguments &a, const char *name, char *&value,
	const char *defaultvalue)
{
	argunit<char *> *p =
		new argunit<char *> (name, value, (char *) defaultvalue);
	a. add (p);
	return;
} /* arg */

template <class type>
inline void arg (arguments &a, const char *name, type *value,
	int &count, int size)
{
	argunit<type> *p = new argunit<type> (name, value, count, size);
	a. add (p);
	return;
} /* arg */

template <class type>
inline void arg (arguments &a, const char *name, type *value,
	int &count, int size, type defaultvalue)
{
	argunit<type> *p = new argunit<type> (name, value, count, size,
		defaultvalue);
	a. add (p);
	return;
} /* argoblig */

template <class type>
inline void argoblig (arguments &arg, const char *name, type &value)
{
	argunit<type> *p = new argunit<type> (name, value);
	p -> set (argflags::obligatory);
	arg. add (p);
	return;
} /* argoblig */

template <class type>
inline void argoblig (arguments &arg, const char *name, type &value,
	type defaultvalue)
{
	argunit<type> *p = new argunit<type> (name, value, defaultvalue);
	p -> set (argflags::obligatory);
	arg. add (p);
	return;
} /* argoblig */

inline void argoblig (arguments &arg, const char *name, char *&value,
	const char *defaultvalue)
{
	argunit<char *> *p =
		new argunit<char *> (name, value, (char *) defaultvalue);
	p -> set (argflags::obligatory);
	arg. add (p);
	return;
} /* argoblig */

template <class type>
inline void argbreak (arguments &arg, const char *name, type &value)
{
	argunit<type> *p = new argunit<type> (name, value);
	p -> set (argflags::breakinterpreting);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argbreak */

template <class type>
inline void argbreak (arguments &arg, const char *name, type &value,
	type defaultvalue)
{
	argunit<type> *p = new argunit<type> (name, value, defaultvalue);
	p -> set (argflags::breakinterpreting);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argbreak */

inline void argbreak (arguments &arg, const char *name, char *&value,
	const char *defaultvalue)
{
	argunit<char *> *p =
		new argunit<char *> (name, value, (char *) defaultvalue);
	p -> set (argflags::breakinterpreting);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argbreak */

inline void argbreak (arguments &arg, const char *name)
{
	char *dummystring = NULL;
	argunit<char *> *p = new argunit<char *> (name, dummystring);
	p -> set (argflags::breakinterpreting);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argbreak */

template <class type>
inline void argswitch (arguments &arg, const char *name, type &value,
	const type &defaultvalue)
{
	argunit<type> *p = new argunit<type> (name, value, defaultvalue);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argswitch */

inline void argswitch (arguments &arg, const char *name, char *&value,
	const char *defaultvalue)
{
	argunit<char *> *p =
		new argunit<char *> (name, value, (char *) defaultvalue);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argswitch */

inline void argswitch (arguments &arg, const char *name)
{
	char *dummystring = NULL;
	argunit<char *> *p = new argunit<char *> (name, dummystring);
	p -> set (argflags::ignorevalue);
	arg. add (p);
	return;
} /* argswitch */

inline void arghelp (arguments &a)
{
	argbreak (a, "?");
	argbreak (a, "h");
	argbreak (a, "H");
	argbreak (a, "-help");
	return;
} /* arghelp */

// --------------------------------------------------

#if defined (_TEXTFILE_H_) || defined (_CAPD_TEXTFILE_H_)

inline void argstreams (arguments &a, char *&logfilename, char *&seqfilename,
	bool &quiet, bool &debug)
{
	arg (a, "-log", logfilename);
	arg (a, "-seq", seqfilename);
	argswitch (a, "-quiet", quiet, true);
	argswitch (a, "-debug", debug, true);
	return;
} /* argstreams */

inline void setstreams (const char *logfilename, char *seqfilename,
	bool quiet, bool debug)
{
	if (debug)
		sbug. show = true;
	if (quiet)
	{
		sout. show = false;
		scon. show = false;
	}
	if (logfilename)
	{
		slog. logfile (logfilename);
		slog. keepforever ();
		sout. logfile (slog);
		serr. logfile (slog);
		if (debug)
			sbug. logfile (slog);
	}
	if (seqfilename)
		sseq. logfile (seqfilename);
#if defined (_TIMEUSED_H_) || defined (_CAPD_TIMEUSED_H_)
	program_time = sout;
#endif
	return;
} /* setstreams */

#ifndef algstreamprepare
#define argstreamprepare(a) \
char *arg_logfilename = NULL; \
char *arg_seqfilename = NULL; \
char *arg_computername = NULL; \
bool arg_quiet = false; \
bool arg_debug = false; \
arg (a, "-comp", arg_computername); \
argstreams (a, arg_logfilename, arg_seqfilename, arg_quiet, arg_debug);
#endif

#ifndef argstreamset
#define argstreamset() \
setstreams (arg_logfilename, arg_seqfilename, arg_quiet, arg_debug); \
slog << commandline (argc, argv) << "\n\n"; \
if (arg_computername) slog << "Running on " << arg_computername << ". "; \
slog << "Start time: " << currenttime () << '\n';
#endif

#endif
/* */

} // namespace homology
} // namespace capd

#endif // _ARG_H_

/// @}
