/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cellbase.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: March 10, 2006.


#ifndef _CAPD_CELLBASE_H_
#define _CAPD_CELLBASE_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubebase.h"
#include "homology/cellmain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace capd {
namespace homology {


// --------------------------------------------------
// ----------- CubicalCell with PointBase -----------
// --------------------------------------------------
// A cubical cell in R^n with edges parallel to the axes
// and with size 0 or 1 in each direction.

template <class coordtype>
class tCellBase
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;

	// the maximal dimension of a cell
	static const int MaxDim = tCubeBase<coordtype>::MaxDim;

	// constructor of an empty cubical cell
	tCellBase ();
	
	// constructor of a cell spanning from one point to another
	tCellBase (const coordtype *c1, const coordtype *c2,
		int spcdim, int celldim = -1);

	// constructor of a cell as an intersection of two cubes
	tCellBase (const tCubeBase<coordtype> &q1,
		const tCubeBase<coordtype> &q2);

	// constructor of a full-dimensional cubical cell
	explicit tCellBase (const tCubeBase<coordtype> &q);

	// copy constructor
	tCellBase (const tCellBase<coordtype> &c);

	// assignment operator
	tCellBase<coordtype> &operator =
		(const tCellBase<coordtype> &c);

	// return the dimension of the cubical cell
	int dim () const;

	// return the dimension of the cubical cell
	operator int () const;

	// return the dimension of the embedding space
	int spacedim () const;

	// return the coordtypes of the left corner of the cell
	const coordtype *leftcoord () const;

	// return the coordinates of the left corner of the cell
	operator const coordtype * () const;
	
	// fill in the given table with the left corner coordinates
	const coordtype *leftcoord (coordtype *c) const;

	// fill in the given table with the right corner coordinates
	const coordtype *rightcoord (coordtype *c = NULL) const;

	// hashing key no. 1 required by the hashing set template
	unsigned hashkey1 () const;

	// hashing key no. 2 required by the hashing set template
	unsigned hashkey2 () const;

	// the name of the objects represented by this class
	static const char *name ();

	// the plural name of the objects represented by this class
	static const char *pluralname ();

	// how to output the cell: as a cartesian product or by two
	// opposite vertices; also, should ' ' be inserted?
	enum OutputBitValues
	{
		BitProduct = 0x01,	// unset => two vertices
		BitSpace = 0x02
	};
	static int OutputBits;

	// friends
	friend inline int operator == (const tCellBase<coordtype> &c1,
		const tCellBase<coordtype> &c2)
	{
		return (c1. n1 == c2. n1) && (c1. n2 == c2. n2);
	} /* operator == */

//	template <class _coordtype>
//	friend int operator == (const tCellBase<_coordtype> &c1,
//		const tCellBase<_coordtype> &c2);

	template <class _coordtype>
	friend int operator != (const tCellBase<_coordtype> &c1,
		const tCellBase<_coordtype> &c2);

private:
	// return the numbers of the cubical cell vertices
	int num1 () const;
	int num2 () const;

	// left vertex: high 6 bits = space dim, the remaining bits
	// form the number of the point in an appropriate pointset
	int n1;

	// right vertex: high 6 bits = cube dim, the remaining bits
	// form the number of the point in an appropriate pointset
	int n2;

	// initialize a new cell given its two corners
	void initialize (const coordtype *c1, const coordtype *c2,
		int spcdim, int celldim = -1);

}; /* class tCellBase */

template<class coordtype>
int tCellBase<coordtype>::OutputBits = 0;

// --------------------------------------------------

template <class coordtype>
inline int tCellBase<coordtype>::num1 () const
{
	return (n1 & NumMask);
} /* tCellBase::num1 */

template <class coordtype>
inline int tCellBase<coordtype>::num2 () const
{
	return (n2 & NumMask);
} /* tCellBase::num2 */

template <class coordtype>
inline void tCellBase<coordtype>::initialize (const coordtype *c1,
	const coordtype *c2, int spcdim, int celldim)
{
	// test the validity of the dimension
	if (spcdim <= 0)
		throw "Non-positive dimension of the space.";
	else if (spcdim >= MaxDim)
		throw "Too high dimension of a cell.";

	// get the number of the left vertex
	n1 = tPointBase<coordtype>::number (c1, spcdim);
	if (n1 < 0)
		throw "Negative number of the left vertex.";

	// get the number of the right vertex
	n2 = tPointBase<coordtype>::number (c2, spcdim);
	if (n2 < 0)
		throw "Negative number of the right vertex.";

	// calculate the dimension of the cubical cell if it is unknown
	if (celldim < 0)
	{
		celldim = 0;
		c1 = tPointBase<coordtype>::coord (n1, spcdim);
		c2 = tPointBase<coordtype>::coord (n2, spcdim);
		for (int i = 0; i < spcdim; i ++)
			if (c2 [i] != c1 [i])
				celldim ++;
	}

	// add the space dimension and the cell dimension to the numbers
	n1 |= (spcdim << NumBits);
	n2 |= (celldim << NumBits);
	return;
} /* tCellBase::initialize */

// --------------------------------------------------

template <class coordtype>
inline tCellBase<coordtype>::tCellBase ()
: n1 (0), n2 (0)
{
	return;
} /* tCellBase::tCellBase */

template <class coordtype>
inline tCellBase<coordtype>::tCellBase
	(const coordtype *c1, const coordtype *c2, int spcdim, int celldim)
{
	initialize (c1, c2, spcdim, celldim);
	return;
} /* tCellBase::tCellBase */

template <class coordtype>
inline tCellBase<coordtype>::tCellBase
	(const tCubeBase<coordtype> &q1, const tCubeBase<coordtype> &q2)
{
	// get the dimension of the space and check for consistency
	int spcdim = q1. dim ();
	if (spcdim != q2. dim ())
		throw "Trying to intersect cubes of different dimension.";

	// get the coordinates of minimal vertices of both cubes
	const coordtype *c1 = q1. coord ();
	const coordtype *c2 = q2. coord ();

	// prepare tables for the new coordinates of the cubical cell
	coordinate left [MaxDim];
	coordinate right [MaxDim];

	// compute the new coordinates of the cubical cell
	int celldim = CommonCell (left, right, c1, c2, spcdim,
		tPointBase<coordtype>::wrap (spcdim));
	
	// create the cell as computed
	initialize (left, right, spcdim, celldim);

	return;
} /* tCellBase::tCellBase */

template <class coordtype>
inline tCellBase<coordtype>::tCellBase (const tCubeBase<coordtype> &q)
: n1 (q. n)
{
	// create the opposite corner of the cell
	int spcdim = n1 >> NumBits;
	const coordtype *c1 = tPointBase<coordtype>::coord (num1 (), spcdim);
	coordtype c2 [MaxDim];
	for (int i = 0; i < spcdim; ++ i)
		c2 [i] = c1 [i] + 1;

	// determine the number of the opposite corner and add cell dimension
	n2 = tPointBase<coordtype>::number (c2, spcdim) |
		(spcdim << NumBits);
	return;
} /* tCellBase::tCellBase */

template <class coordtype>
inline tCellBase<coordtype>::tCellBase (const tCellBase<coordtype> &q)
: n1 (q. n1), n2 (q. n2)
{
	return;
} /* tCellBase::tCellBase */

template <class coordtype>
inline tCellBase<coordtype> &tCellBase<coordtype>::operator =
	(const tCellBase<coordtype> &q)
{
	n1 = q. n1;
	n2 = q. n2;
	return *this;
} /* tCellBase::operator = */

template <class coordtype>
inline int tCellBase<coordtype>::dim () const
{
	return (n2 >> NumBits);
} /* tCellBase::dim */

template <class coordtype>
inline tCellBase<coordtype>::operator int () const
{
	return dim ();
} /* tCellBase::operator int */

template <class coordtype>
inline int tCellBase<coordtype>::spacedim () const
{
	return (n1 >> NumBits);
} /* tCellBase::spacedim */

template <class coordtype>
inline const coordtype *tCellBase<coordtype>::leftcoord () const
{
	return tPointBase<coordtype>::coord (num1 (), spacedim ());
} /* tCellBase::leftcoord */

template <class coordtype>
inline tCellBase<coordtype>::operator const coordtype * () const
{
	return leftcoord ();
} /* tCellBase::operator coordtype * */

template <class coordtype>
inline const coordtype *tCellBase<coordtype>::leftcoord
	(coordtype *c) const
{
	const coordtype *leftc = tPointBase<coordtype>::coord (num1 (),
		spacedim ());
	if (c)
		copycoord (c, leftc, spacedim ());
	return leftc;
} /* tCellBase::leftcoord */

template <class coordtype>
inline const coordtype *tCellBase<coordtype>::rightcoord
	(coordtype *c) const
{
	const coordtype *rightc = tPointBase<coordtype>::coord (num2 (),
		spacedim ());
	if (c)
		copycoord (c, rightc, spacedim ());
	return rightc;
} /* tCellBase::rightcoord */

template <class coordtype>
inline unsigned tCellBase<coordtype>::hashkey1 () const
{
	return ((n1 & 0x555555) << 17) + ((n1 & 0xAA00AA) << 7) +
		((n2 & 0xAA00) >> 7);
} /* tCellBase::hashkey1 */

template <class coordtype>
inline unsigned tCellBase<coordtype>::hashkey2 () const
{
	return ((n2 & 0xAAAAAA) << 18) + ((n2 & 0x555555) << 10) +
		((n1 & 0x55555500) >> 8);
} /* tCellBase::hashkey2 */

template <class coordtype>
const char *tCellBase<coordtype>::name ()
{
	return "cubical cell";
} /* tCellBase::name */

template <class coordtype>
const char *tCellBase<coordtype>::pluralname ()
{
	return "cubical cells";
} /* tCellBase::pluralname */

// --------------------------------------------------

/*
template <class coordtype>
inline int operator == (const tCellBase<coordtype> &c1,
	const tCellBase<coordtype> &c2)
{
	return (c1. n1 == c2. n1) && (c1. n2 == c2. n2);
}*/ /* operator == */

template <class coordtype>
inline int operator != (const tCellBase<coordtype> &c1,
	const tCellBase<coordtype> &c2)
{
	return (c1. n1 != c2. n1) || (c1. n2 != c2. n2);
} /* operator != */

// --------------------------------------------------

// Compute the Cartesian product of two cells.
template <class coordtype>
inline tCellBase<coordtype> operator *
	(const tCellBase<coordtype> &c1,
	const tCellBase<coordtype> &c2)
{
	const coordtype *c1left = c1. leftcoord ();
	const coordtype *c2left = c2. leftcoord ();
	const coordtype *c1right = c1. rightcoord ();
	const coordtype *c2right = c2. rightcoord ();
	int d1 = c1. spacedim (), d2 = c2. spacedim ();
	if (d1 + d2 >= tCellBase<coordtype>::MaxDim)
		throw "Too high dimension of a Cartesian product of cells.";
	coordtype lcoord [tCellBase<coordtype>::MaxDim];
	coordtype rcoord [tCellBase<coordtype>::MaxDim];
	int j = 0;
	int i;
	for (i = 0; i < d1; i ++)
	{
		lcoord [j] = c1left [i];
		rcoord [j] = c1right [i];
		j ++;
	}
	for (i = 0; i < d2; i ++)
	{
		lcoord [j] = c2left [i];
		rcoord [j] = c2right [i];
		j ++;
	}
	return tCellBase<coordtype> (lcoord, rcoord, d1 + d2,
		c1. dim () + c2. dim ());
} /* operator * */

// Write a cell to an output stream.
template <class coordtype>
inline std::ostream &operator << (std::ostream &out,
	const tCellBase<coordtype> &c)
{
	return WriteCubicalCell (out, c);
} /* operator << */

// Read a cell from an input stream.
template <class coordtype>
inline std::istream &operator >> (std::istream &in,
	tCellBase<coordtype> &c)
{
	return ReadCubicalCell (in, c);
} /* operator >> */

// --------------------------------------------------

// a boundary element of a cell; if only existing cells are considered,
// returns '*this' if the requested boundary cell doesn't exist
template <class coordtype>
inline tCellBase<coordtype> boundarycell (const tCellBase<coordtype> &q,
	int i, bool onlyexisting)
{
	return CubicalBoundaryCell (q, i, onlyexisting);
} /* boundarycell */

// a boundary of a cubical cell - including non-existing cells
template <class coordtype>
inline tCellBase<coordtype> boundarycell (const tCellBase<coordtype> &q,
	int i)
{
	return CubicalBoundaryCell (q, i);
} /* boundarycell */
	
// the length of the boundary of a cell
template <class coordtype>
inline int boundarylength (const tCellBase<coordtype> &q)
{
	return CubicalBoundaryLength (q);
} /* boundarylength */

// a coefficient in the boundary of a cell
template <class coordtype>
inline int boundarycoef (const tCellBase<coordtype> &q, int i)
{
	return CubicalBoundaryCoef (q, i);
} /* boundarycoef */

} // namespace homology
} // namespace capd


#endif // _CELLBASE_H_

/// @}
