/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cellfix.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: March 10, 2006.


#ifndef _CAPD_CELLFIX_H_
#define _CAPD_CELLFIX_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubefix.h"
#include "homology/cellmain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace capd {
namespace homology {


// --------------------------------------------------
// ----------- CubicalCell with fixed dim -----------
// --------------------------------------------------
// A cubical cell in R^n with edges parallel to the axes
// and with size 0 or 1 in each direction.

template <int dimfix, class coordtype>
class tCellFix
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;

	// the maximal dimension of a cell (unused)
	static const int MaxDim = tCubeFix<dimfix,coordtype>::MaxDim;

	// constructor of an empty cubical cell
	tCellFix ();
	
	// constructor of a cell spanning from one point to another
	tCellFix (const coordtype *c1, const coordtype *c2,
		int spcdim = 0, int celldim = -1);

	// constructor of a cell as an intersection of two cubes
	tCellFix (const tCubeFix<dimfix,coordtype> &q1,
		const tCubeFix<dimfix,coordtype> &q2);

	// constructor of a full-dimensional cubical cell
	explicit tCellFix (const tCubeFix<dimfix,coordtype> &q);

	// copy constructor
	tCellFix (const tCellFix<dimfix,coordtype> &c);

	// assignment operator
	tCellFix<dimfix,coordtype> &operator =
		(const tCellFix<dimfix,coordtype> &c);

	// return the dimension of the cubical cell
	int dim () const;

	// return the dimension of the cubical cell
	operator int () const;

	// return the dimension of the embedding space
	int spacedim () const;

	// return the coordtypes of the left corner of the cell
	const coordtype *leftcoord () const;

	// return the coordinates of the left corner of the cell
	operator const coordtype * () const;
	
	// fill in the given table with the left corner coordinates
	const coordtype *leftcoord (coordtype *c) const;

	// fill in the given table with the right corner coordinates
	const coordtype *rightcoord (coordtype *c = NULL) const;

	// hashing key no. 1 required by the hashing set template
	unsigned hashkey1 () const;

	// hashing key no. 2 required by the hashing set template
	unsigned hashkey2 () const;

	// the name of the objects represented by this class
	static const char *name ();

	// the plural name of the objects represented by this class
	static const char *pluralname ();

	// how to output the cell: as a cartesian product or by two
	// opposite vertices; also, should ' ' be inserted?
	enum OutputBitValues
	{
		BitProduct = 0x01,	// unset => two vertices
		BitSpace = 0x02
	};
	static int OutputBits;

	// friends
	template <int _dimfix, class _coordtype>
	friend int operator ==
		(const tCellFix<_dimfix,_coordtype> &c1,
		const tCellFix<_dimfix,_coordtype> &c2);
	template <int _dimfix, class _coordtype>
	friend int operator !=
		(const tCellFix<_dimfix,_coordtype> &c1,
		const tCellFix<_dimfix,_coordtype> &c2);

private:
	// a table with the left coordinate of the cell
	coordtype tab [dimfix];

	// high 6 bits = the dimension of the cell (not: space);
	// remaining bits set to 1 iff the cell has width 1
	// in the specified direction
	int n;

	// initialize a new cell given its two corners
	void initialize (const coordtype *c1, const coordtype *c2,
		int celldim = -1);

}; /* class tCellFix */

template <int dimfix, class coordtype>
int tCellFix<dimfix,coordtype>::OutputBits = 0;

// --------------------------------------------------

template <int dimfix, class coordtype>
inline void tCellFix<dimfix,coordtype>::initialize (const coordtype *c1,
	const coordtype *c2, int celldim)
{
	// copy the left corner coordinates to the cell
	tPointBase<coordtype>::wrapcopy (tab, c1, dimfix);

	// prepare wrapped right coordinates for the detection of cell bits
	coordtype r [dimfix];
	tPointBase<coordtype>::wrapcopy (r, c2, dimfix);

	// if the dimension of the cell is known, then compute the bits only
	if (celldim >= 0)
	{
		n = celldim << NumBits;
		if (!celldim)
			return;
		for (int i = 0; i < dimfix; ++ i)
			if (tab [i] != r [i])
				n |= (1 << i);
		return;
	}

	// set the bits of the cell and compute its actual dimension
	n = 0;
	int dim = 0;
	for (int i = 0; i < dimfix; ++ i)
		if (tab [i] != r [i])
		{
			n |= (1 << i);
			++ dim;
		}
	n |= dim << NumBits;

	return;
} /* tCellFix::initialize */

// --------------------------------------------------

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::tCellFix ()
:n (0)
{
	return;
} /* tCellFix::tCellFix */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::tCellFix
	(const coordtype *c1, const coordtype *c2, int spcdim, int celldim)
{
	// test the validity of the dimension
	if (spcdim < 0)
		throw "Negative dimension of the space.";
	else if ((spcdim > 0) && (spcdim != dimfix))
		throw "Wrong dimension of a cell.";

	// initialize the cell
	initialize (c1, c2, celldim);
	tPointBase<coordtype>::outOfControl = true;
	return;
} /* tCellFix::tCellFix */
	
template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::tCellFix
	(const tCubeFix<dimfix,coordtype> &q1,
	const tCubeFix<dimfix,coordtype> &q2)
{
	// get the coordinates of minimal vertices of both cubes
	const coordtype *c1 = q1. coord ();
	const coordtype *c2 = q2. coord ();

	// prepare tables for the new coordinates of the cubical cell
	coordinate left [dimfix];
	coordinate right [dimfix];

	// compute the new coordinates of the cubical cell
	int celldim = CommonCell (left, right, c1, c2, dimfix,
		tPointBase<coordtype>::wrap (dimfix));
	
	// create the cell as computed
	initialize (left, right, celldim);

	return;
} /* tCellFix::tCellFix */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::tCellFix
	(const tCubeFix<dimfix,coordtype> &q)
: n ((dimfix << NumBits) | NumMask)
{
	return;
} /* tCellFix::tCellFix */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::tCellFix
	(const tCellFix<dimfix,coordtype> &q)
: n (q. n)
{
	copycoord (tab, q. tab, dimfix);
	return;
} /* tCellFix::tCellFix */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype> &tCellFix<dimfix,coordtype>::operator =
	(const tCellFix<dimfix,coordtype> &q)
{
	memcpy (tab, q. tab, dimfix * sizeof (coordtype));
	n = q. n;
	return *this;
} /* tCellFix::operator = */

template <int dimfix, class coordtype>
inline int tCellFix<dimfix,coordtype>::dim () const
{
	return (int) (n >> NumBits);
} /* tCellFix::dim */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::operator int () const
{
	return dim ();
} /* tCellFix::operator int */

template <int dimfix, class coordtype>
inline int tCellFix<dimfix,coordtype>::spacedim () const
{
	return dimfix;
} /* tCellFix::spacedim */

template <int dimfix, class coordtype>
inline const coordtype *tCellFix<dimfix,coordtype>::leftcoord () const
{
	return tab;
} /* tCellFix::leftcoord */

template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype>::operator const coordtype * () const
{
	return leftcoord ();
} /* tCellFix::operator coordtype * */

template <int dimfix, class coordtype>
inline const coordtype *tCellFix<dimfix,coordtype>::leftcoord
	(coordtype *c) const
{
	if (c)
	{
		for (int i = 0; i < dimfix; ++ i)
			c [i] = tab [i];
	}
	return tab;
} /* tCellFix::leftcoord */

template <int dimfix, class coordtype>
inline const coordtype *tCellFix<dimfix,coordtype>::rightcoord
	(coordtype *c) const
{
	if (!c)
		return NULL;
	for (int i = 0; i < dimfix; ++ i)
		c [i] = tab [i] + 1;
	tPointBase<coordtype>::wrapcoord (c, spacedim ());
	return c;
} /* tCellFix::rightcoord */

template <int dimfix, class coordtype>
inline unsigned tCellFix<dimfix,coordtype>::hashkey1 () const
{
	switch (dimfix)
	{
	case 1:
		return (unsigned) ((tab [0]) << 12) ^ n;
	case 2:
		return ((((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6)) ^ n;
	default:
		return ((((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6) +
			(((unsigned) (tab [2])) >> 6)) ^ n;
	}
} /* tCellFix::hashkey1 */

template <int dimfix, class coordtype>
inline unsigned tCellFix<dimfix,coordtype>::hashkey2 () const
{
	switch (dimfix)
	{
	case 1:
		return ((unsigned) (tab [0]) << 3) ^ (n << 5);
	case 2:
		return (((unsigned) (tab [0]) >> 1) +
			((unsigned) (tab [1]) << 13)) ^ (n << 5);
	default:
		return ((((unsigned) (tab [dimfix - 1])) << 20) +
			(((unsigned) (tab [dimfix - 2])) << 9) +
			(((unsigned) (tab [dimfix - 3])) >> 1)) ^ (n << 5);
	}
} /* tCellFix::hashkey2 */

template <int dimfix, class coordtype>
const char *tCellFix<dimfix,coordtype>::name ()
{
	return "cubical cell";
} /* tCellFix::name */

template <int dimfix, class coordtype>
const char *tCellFix<dimfix,coordtype>::pluralname ()
{
	return "cubical cells";
} /* tCellFix::pluralname */

// --------------------------------------------------

template <int dimfix, class coordtype>
inline int operator == (const tCellFix<dimfix,coordtype> &c1,
	const tCellFix<dimfix,coordtype> &c2)
{
	return ((c1. n == c2. n) && (!memcmp (c1. tab, c2. tab,
		dimfix * sizeof (coordtype))));
} /* operator == */

template <int dimfix, class coordtype>
inline int operator != (const tCellFix<dimfix,coordtype> &c1, const tCellFix<dimfix,coordtype> &c2)
{
	return ((c1. n != c2. n) || (memcmp (c1. tab, c2. tab,
		dimfix * sizeof (coordtype))));
} /* operator != */

// --------------------------------------------------

template <int dim1, int dim2, class coordtype>
inline tCellFix<dim1+dim2,coordtype> operator *
	(const tCellFix<dim1,coordtype> &c1,
	const tCellFix<dim2,coordtype> &c2)
// Compute the Cartesian product of two cells.
{
	const coordtype *c1left = c1. leftcoord ();
	const coordtype *c2left = c2. leftcoord ();
	coordtype c1right [dim1], c2right [dim2];
	c1. rightcoord (c1right);
	c2. rightcoord (c2right);
	coordtype lcoord [dim1 + dim2], rcoord [dim1 + dim2];
	int j = 0;
	int i;
	for (i = 0; i < dim1; ++ i)
	{
		lcoord [j] = c1left [i];
		rcoord [j] = c1right [i];
		j ++;
	}
	for (i = 0; i < dim2; ++ i)
	{
		lcoord [j] = c2left [i];
		rcoord [j] = c2right [i];
		j ++;
	}
	return tCellFix<dim1+dim2,coordtype> (lcoord, rcoord,
		dim1 + dim2, c1. dim () + c2. dim ());
} /* operator * */

// Write a cell to an output stream.
template <int dimfix, class coordtype>
inline std::ostream &operator << (std::ostream &out,
	const tCellFix<dimfix,coordtype> &c)
{
	return WriteCubicalCell (out, c);
} /* operator << */

// Read a cell from an input stream.
template <int dimfix, class coordtype>
inline std::istream &operator >> (std::istream &in,
	tCellFix<dimfix,coordtype> &c)
{
	return ReadCubicalCell (in, c);
} /* operator >> */

// --------------------------------------------------

// a boundary element of a cell; if only existing cells are considered,
// returns '*this' if the requested boundary cell doesn't exist
template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype> boundarycell (const tCellFix<dimfix,coordtype> &q,
	int i, bool onlyexisting)
{
	return CubicalBoundaryCell (q, i, onlyexisting);
} /* boundarycell */

// a boundary of a cubical cell - including non-existing cells
template <int dimfix, class coordtype>
inline tCellFix<dimfix,coordtype> boundarycell (const tCellFix<dimfix,coordtype> &q,
	int i)
{
	return CubicalBoundaryCell (q, i);
} /* boundarycell */
	
// the length of the boundary of a cell
template <int dimfix, class coordtype>
inline int boundarylength (const tCellFix<dimfix,coordtype> &q)
{
	return CubicalBoundaryLength (q);
} /* boundarylength */

// a coefficient in the boundary of a cell
template <int dimfix, class coordtype>
inline int boundarycoef (const tCellFix<dimfix,coordtype> &q, int i)
{
	return CubicalBoundaryCoef (q, i);
} /* boundarycoef */

} // namespace homology
} // namespace capd


#endif // _CELLFIX_H_

/// @}
