/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cellvar.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: March 10, 2006.


#ifndef _CAPD_CELLVAR_H_
#define _CAPD_CELLVAR_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cubevar.h"
#include "homology/cellmain.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace capd {
namespace homology {


// --------------------------------------------------
// ------- CubicalCell with allocated memory --------
// --------------------------------------------------
// A cubical cell in R^n with edges parallel to the axes
// and with size 0 or 1 in each direction.

template <class coordtype>
class tCellVar
{
public:
	// the type of the coordinates
	typedef coordtype CoordType;

	// the maximal dimension of a cell
	static const int MaxDim = (MaxBasDim < NumBits) ? MaxBasDim : NumBits;

	// various constructors: an empty cub. cell, a cell spanning
	// from one point to another, the intersection of two cubes
	// or just a full-dimensional cubical cell
	tCellVar ();
	tCellVar (const coordtype *c1, const coordtype *c2,
		int spcdim, int celldim = -1);
	tCellVar (const tCubeVar<coordtype> &q1,
		const tCubeVar<coordtype> &q2);
	explicit tCellVar (const tCubeVar<coordtype> &q);

	// copy constructor
	tCellVar (const tCellVar<coordtype> &c);

	// assignment operator
	tCellVar<coordtype> &operator =
		(const tCellVar<coordtype> &c);

	// return the dimension of the cubical cell
	int dim () const;

	// return the dimension of the cubical cell
	operator int () const;

	// return the dimension of the embedding space of the cubical cell
	int spacedim () const;

	// return the coordinates of the left corner of the cubical cell
	const coordtype *leftcoord () const;

	// return the coordinates of the left corner of the cubical cell
	operator const coordtype * () const;
	
	// fill in the given table with the right corner coordinates
	const coordtype *leftcoord (coordtype *c) const;

	// fill in the given table with the right corner coordinates
	const coordtype *rightcoord (coordtype *c) const;

	// two hashing keys required by the hashing set template
	unsigned hashkey1 () const;
	unsigned hashkey2 () const;

	// the name of the object(s) represented by this class
	static const char *name ();
	static const char *pluralname ();

	// how to output the cell: as a cartesian product or by two
	// opposite vertices; also, should the space be added
	enum OutputBitValues
	{
		BitProduct = 0x01,	// unset => two vertices
		BitSpace = 0x02
	};
	static int OutputBits;

	// friends
	template <class _coordtype>
	friend int operator == (const tCellVar<_coordtype> &c1,
		const tCellVar<_coordtype> &c2);

private:
	// the lower left etc. corner of the cell
	coordtype *tab;

	// high 6 bits = the dimension of the space;
	// remaining bits set to 1 iff the cell has width 1
	// in the specified direction
	int n;

	// initialize a new cell given its two corners
	void initialize (const coordtype *c1, const coordtype *c2,
		int spcdim, int celldim = -1);

}; /* class tCellVar */

template<class coordtype>
int tCellVar<coordtype>::OutputBits = 0;

// --------------------------------------------------

template <class coordtype>
inline void tCellVar<coordtype>::initialize (const coordtype *c1,
	const coordtype *c2, int spcdim, int)
{
	// test the validity of the dimension
	if (spcdim <= 0)
		throw "Non-positive dimension of the space.";
	else if (spcdim >= MaxDim)
		throw "Too high dimension of a cell.";

	// initialize the internal number with the space dimension
	n = (spcdim << NumBits);

	// prepare a table for coordinates if necessary
	tab = new coordtype [spcdim];
	if (!tab)
		throw "Not enough memory to create a cell.";

	// copy the left corner coordinates to the cell
	tPointBase<coordtype>::wrapcopy (tab, c1, spcdim);

	// prepare wrapped right coordinates for the detection of cell bits
	coordtype *r = new coordtype [MaxDim];
	tPointBase<coordtype>::wrapcopy (r, c2, spcdim);

	// set the bits of the cell
	for (int i = 0; i < spcdim; i ++)
		if (tab [i] != r [i])
			n |= (1 << i);
	delete [] r;

	return;
} /* tCellVar::initialize */

template <class coordtype>
inline tCellVar<coordtype>::tCellVar ()
: tab (NULL), n (0)
{
	return;
} /* tCellVar::tCellVar */

template <class coordtype>
inline tCellVar<coordtype>::tCellVar
	(const coordtype *c1, const coordtype *c2, int spcdim, int celldim)
{
	initialize (c1, c2, spcdim, celldim);
	tPointBase<coordtype>::outOfControl = true;
	return;
} /* tCellVar::tCellVar */

template <class coordtype>
inline int tCellVar<coordtype>::dim () const
{
	int count = 0;
	for (int i = 0; i < NumBits; i ++)
		if (n & (1 << i))
			count ++;
	return count;
} /* tCellVar::dim */

template <class coordtype>
inline tCellVar<coordtype>::operator int () const
{
	return dim ();
} /* operator int */

template <class coordtype>
inline int tCellVar<coordtype>::spacedim () const
{
	return (int) (n >> NumBits);
} /* tCellVar::spacedim */

template <class coordtype>
inline tCellVar<coordtype>::tCellVar (const tCellVar<coordtype> &q)
{
	n = q. n;
	tab = new coordtype [spacedim ()];
	if (!tab)
		throw "Not enough memory to copy a cubical cell.";
	copycoord (tab, q. tab, spacedim ());
	return;
} /* tCellVar::tCellVar */

template <class coordtype>
inline tCellVar<coordtype> &tCellVar<coordtype>::operator =
	(const tCellVar<coordtype> &q)
{
	// determine the new dimension
	int d = q. spacedim ();

	// allocate a new table if the dimension is to be changed
	if (d != spacedim ())
	{
		if (tab)
			delete [] tab;
		tab = new coordtype [d];
		if (!tab)
			throw "Not enough memory to copy a cubical cell.";
	}

	// copy the data of the cubical cell
	copycoord (tab, q. tab, d);
	n = q. n;

	return *this;
} /* tCellVar::operator = */

template <class coordtype>
inline const coordtype *tCellVar<coordtype>::leftcoord () const
{
	return tab;
} /* tCellVar::leftcoord */

template <class coordtype>
inline tCellVar<coordtype>::operator const coordtype * () const
{
	return leftcoord ();
} /* tCellVar::operator coordtype * */

template <class coordtype>
inline const coordtype *tCellVar<coordtype>::leftcoord (coordtype *c)
	const
{
	int d = spacedim ();
	if (c)
	{
		for (int i = 0; i < d; i ++)
			c [i] = tab [i];
	}
	return tab;
} /* tCellVar::leftcoord */

template <class coordtype>
inline const coordtype *tCellVar<coordtype>::rightcoord (coordtype *c)
	const
{
	const coordtype *left = leftcoord ();
	int d = spacedim ();
	for (int i = 0; i < d; i ++)
	{
		c [i] = left [i];
		if (n & (1 << i))
			++ (c [i]);
	}
	tPointBase<coordtype>::wrapcoord (c, d);
	return c;
} /* tCellVar::rightcoord */

template <class coordtype>
inline unsigned tCellVar<coordtype>::hashkey1 () const
{
	switch (spacedim ())
	{
	case 0:
		return 0;
	case 1:
		return ((unsigned) (tab [0]) << 12) ^ n;
	case 2:
		return ((((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6)) ^ n;
	default:
		return ((((unsigned) (tab [0])) << 18) +
			(((unsigned) (tab [1])) << 6) +
			(((unsigned) (tab [2])) >> 6)) ^ n;
	}
} /* tCellVar::hashkey1 */

template <class coordtype>
inline unsigned tCellVar<coordtype>::hashkey2 () const
{
	int d = spacedim ();
	switch (d)
	{
	case 0:
		return 0;
	case 1:
		return ((unsigned) (tab [0]) << 3) ^ (n << 5);
	case 2:
		return (((unsigned) (tab [0]) >> 1) +
			((unsigned) (tab [1]) << 13)) ^ (n << 5);
	default:
		return ((((unsigned) (tab [d - 1])) << 20) +
			(((unsigned) (tab [d - 2])) << 9) +
			(((unsigned) (tab [d - 3])) >> 1)) ^ (n << 5);
	}
} /* tCellVar::hashkey2 */

template <class coordtype>
const char *tCellVar<coordtype>::name ()
{
	return "cubical cell";
} /* tCellVar::name */

template <class coordtype>
const char *tCellVar<coordtype>::pluralname ()
{
	return "cubical cells";
} /* tCellVar::pluralname */

// --------------------------------------------------

template <class coordtype>
inline int operator == (const tCellVar<coordtype> &c1,
	const tCellVar<coordtype> &c2)
{
	if (c1. n != c2. n)
		return 0;
	if (!c1. n)
		return 0;
	return thesame (c1. tab, c2. tab, c1. spacedim ());
} /* operator == */

template <class coordtype>
inline int operator != (const tCellVar<coordtype> &c1, const tCellVar<coordtype> &c2)
{
	return !(c1 == c2);
} /* operator != */

// --------------------------------------------------

template <class coordtype>
inline tCellVar<coordtype>::tCellVar
	(const tCubeVar<coordtype> &q1, const tCubeVar<coordtype> &q2)
{
	// prepare tables for coordinates of the cubical cell
	coordtype *left = new coordtype [MaxDim];
	coordtype *right = new coordtype [MaxDim];

	// get the coordinates of minimal vertices of both cubes
	const coordtype *c1 = q1;
	const coordtype *c2 = q2;

	// get the dimension of the space and check for consistency
	int spcdim = (int) q1;
	if (spcdim != (int) q2)
		throw "Trying to intersect cubes of different dimension.";

	// calculate the coordinates of both vertices of the cubical cell
	// and the dimension of the cubical cell
	int celldim = 0;
	const coordtype *wrap = tPointBase<coordtype>::wrap (spcdim);
	for (int i = 0; i < spcdim; i ++)
	{
		if (c1 [i] == c2 [i])
		{
			left [i] = c1 [i];
			right [i] = c1 [i];
			++ right [i];
			++ celldim;
		}
		else if ((c1 [i] - c2 [i] == -1) || (wrap && wrap [i] &&
			(c1 [i] - c2 [i] == wrap [i] - 1)))
		{
			left [i] = c2 [i];
			right [i] = c2 [i];
		}
		else if ((c1 [i] - c2 [i] == 1) || (wrap && wrap [i] &&
			(c1 [i] - c2 [i] == -wrap [i] + 1)))
		{
			left [i] = c1 [i];
			right [i] = c1 [i];
		}
		else
			throw "The cubes do not intersect.";
	}

	// initialize the data of the cube
	initialize (left, right, spcdim, celldim);

	delete [] right;
	delete [] left;
	return;
} /* tCellVar::tCellVar */

template <class coordtype>
inline tCellVar<coordtype>::tCellVar (const tCubeVar<coordtype> &q)
{
	// get the coordinates of minimal vertex of the cube
	const coordtype *left = q;

	// prepare a table for coordinates of the other vertex of the cell
	coordtype *right = new coordtype [MaxDim];

	// get the dimension of the space and of the cell
	int d = (int) q;

	// calculate the coordinates of other vertex of the cell
	for (int i = 0; i < d; i ++)
	{
		right [i] = left [i];
		right [i] ++;
	}

	initialize (left, right, d, d);

	delete [] right;
	return;
} /* tCellVar::tCellVar */

// --------------------------------------------------

template <class coordtype>
inline tCellVar<coordtype> operator *
	(const tCellVar<coordtype> &c1,
	const tCellVar<coordtype> &c2)
// Compute the Cartesian product of two cells.
{
	const coordtype *c1left = c1. leftcoord ();
	const coordtype *c2left = c2. leftcoord ();
	coordtype c1right [tCellVar<coordtype>::MaxDim];
	coordtype c2right [tCellVar<coordtype>::MaxDim];
	c1. rightcoord (c1right);
	c2. rightcoord (c2right);
	int d1 = c1. spacedim (), d2 = c2. spacedim ();
	if (d1 + d2 >= tCellVar<coordtype>::MaxDim)
		throw "Too high dimension of a Cartesian product of cells.";
	coordtype lcoord [tCellVar<coordtype>::MaxDim];
	coordtype rcoord [tCellVar<coordtype>::MaxDim];
	int j = 0;
	int i;
	for (i = 0; i < d1; i ++)
	{
		lcoord [j] = c1left [i];
		rcoord [j] = c1right [i];
		j ++;
	}
	for (i = 0; i < d2; i ++)
	{
		lcoord [j] = c2left [i];
		rcoord [j] = c2right [i];
		j ++;
	}
	return tCellVar<coordtype> (lcoord, rcoord, d1 + d2,
		c1. dim () + c2. dim ());
} /* operator * */

// Write a cell to an output stream.
template <class coordtype>
inline std::ostream &operator << (std::ostream &out,
	const tCellVar<coordtype> &c)
{
	return WriteCubicalCell (out, c);
} /* operator << */

// Read a cell from an input stream.
template <class coordtype>
inline std::istream &operator >> (std::istream &in,
	tCellVar<coordtype> &c)
{
	return ReadCubicalCell (in, c);
} /* operator >> */

// --------------------------------------------------

// a boundary element of a cell; if only existing cells are considered,
// returns '*this' if the requested boundary cell doesn't exist
template <class coordtype>
inline tCellVar<coordtype> boundarycell (const tCellVar<coordtype> &q,
	int i, bool onlyexisting)
{
	return CubicalBoundaryCell (q, i, onlyexisting);
} /* boundarycell */

// a boundary of a cubical cell - including non-existing cells
template <class coordtype>
inline tCellVar<coordtype> boundarycell (const tCellVar<coordtype> &q,
	int i)
{
	return CubicalBoundaryCell (q, i);
} /* boundarycell */
	
// the length of the boundary of a cell
template <class coordtype>
inline int boundarylength (const tCellVar<coordtype> &q)
{
	return CubicalBoundaryLength (q);
} /* boundarylength */

// a coefficient in the boundary of a cell
template <class coordtype>
inline int boundarycoef (const tCellVar<coordtype> &q, int i)
{
	return CubicalBoundaryCoef (q, i);
} /* boundarycoef */

} // namespace homology
} // namespace capd


#endif // _CELLVAR_H_

/// @}
