/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubisets.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: April 27, 2007.


#ifndef _CAPD_CUBISETS_H_
#define _CAPD_CUBISETS_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/pointbas.h"
#include "homology/cube.h"
#include "homology/cell.h"
#include "homology/cubacycl.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace capd {
namespace homology {


// --------------------------------------------------
// ------------------- BitFields --------------------
// --------------------------------------------------
// A table of bit fields for storing the information
// on the verified combinations of cubes' neighbors.

class BitFields
{
public:
	// default constructor
	BitFields ();

	// default destructor
	~BitFields ();

	// set the limit for the number of kbytes used for BitFields;
	// limit applied to dimension < 0 is copied to all the
	// others and can be retrieved as the limit for dimension 0;
	// limit for dim. 0 is the default if not defined otherwise
	void setkblimit (int limit = -1, int dim = -1);

	// get the current limit for the given dimension
	int getkblimit (int dim = 0);

	// get the corresponding bit field set and allocate it if not
	// used so far; if no bit field set is in use, return 0
	SetOfBitFields *operator [] (int dim);

	// forget the given bit field set or all the sets
	void forget (int dim = -1);

protected:
	// the table of bit field sets for each dimension
	SetOfBitFields **tab;

	// the length of the currently allocated table
	int len;

	// memory limit for each table (in kilobytes); -1 for none
	int *maxkb;

	// extend the table to the given length if necessary
	void extend (int n);

	// allocate the specific bit field set
	void allocate (int dim);

}; /* class BitFields */

inline BitFields::BitFields ()
{
	tab = NULL;
	len = 0;
	maxkb = NULL;
	return;
} /* BitFields::BitFields */

inline SetOfBitFields *BitFields::operator [] (int dim)
{
	// make sure the dimension is positive
	if (dim <= 0)
		throw "Trying to get a bit field set of non-positive dim.";

	// extend the tables if necessary
	if (dim >= len)
		extend (dim + 1);

	// allocate the bit field set if necessary
	if (!tab [dim] && maxkb [dim])
		allocate (dim);

	// return the pointer to the requested bit field set
	return tab [dim];
} /* BitFields::operator [] */

// the global table of BitFields which store the acyclicity information
extern BitFields knownbits;


// --------------------------------------------------

/// tabulated configurations of neighbors for various dimensions.
class Tabulated
{
public:
	/// the default constructor
	Tabulated ();

	/// the destructor
	~Tabulated ();

	/// read tabulated configurations from a file
	int read (int dim, const char *filename);

	/// write tabulated configurations to a file
	int write (int dim, const char *filename) const;

	/// compute tabulated configurations for a specific dimension
	int compute (int dim);

	/// set tabulated configuration bits to a given table
	/// (this buffer will not be deallocated automatically);
	/// set to the null pointer to disable the table
	int define (int dim, char *buffer);

	/// retrieve the buffer allocated for the specific dimension
	/// or return the null pointer if none
	const char *operator [] (int dim) const;

	/// retrieve the given bit from the given table
	static int get (const char *table, int bitnumber);

	/// set the given bit in the given table
	static void set (char *table, int bitnumber);

private:
	/// the strict upper bound for the supported dimensions
	static const int maxdim = 8;

	/// the tabulated configurations
	char *tables [maxdim];

	/// should the configuration tables be deallocated?
	bool deallocate [maxdim];

	/// the size of the table in bytes for each dimension
	int size [maxdim];
	
}; /* Tabulated */

// --------------------------------------------------

inline int Tabulated::define (int dim, char *buffer)
{
	if ((dim <= 0) || (dim >= maxdim))
		return -1;
	if (tables [dim] && deallocate [dim])
	{
		delete [] (tables [dim]);
		deallocate [dim] = false;
	}
	tables [dim] = buffer;
	return 0;
} /* Tabulated::define */

inline const char *Tabulated::operator [] (int dim) const
{
	if ((dim <= 0) || (dim >= maxdim))
		return 0;
	else
		return tables [dim];
} /* Tabulated::operator [] */

inline int Tabulated::get (const char *table, int bitnumber)
{
	if (table [bitnumber >> 3] & (1 << (bitnumber & 0x07)))
		return 1;
	else
		return 0;
} /* Tabulated::get */

inline void Tabulated::set (char *table, int bitnumber)
{
	table [bitnumber >> 3] |= (char) (1 << (bitnumber & 0x07));
	return;
} /* Tabulated::set */

// --------------------------------------------------

extern Tabulated tabulated;

// --------------------------------------------------

extern int MaxBddDim;
extern int MaxBddDimPossible;

extern bool UseMalandainCode;

// --------------------------------------------------

template <class euclidom, class tCell>
std::ostream &writegenerators (std::ostream &out, const chain<euclidom> *hom,
	const chaincomplex<euclidom> &c,
	const gcomplex<tCell,euclidom> &g, const int *level, int xdim,
	int format = 0)
// Write projected homology generators of the cubical complex to a file.
// Note: The main loop(s) are copied from the function "createprojection".
// Format: 0 = text format, 1 = chl format.
{
	bool firstlist = true;
	for (int d = 0; d <= c. dim (); d ++)
		if ((!level || level [d]) && ((int) hom [d]))
		{
			if (firstlist)
				firstlist = false;
			else if (format == 0)
				out << '\n';
			if (format == 0)
				if ((int) hom [d] == 1)
					out << "The generator of H_" << d <<
						" follows:" << '\n';
				else
					out << "The " << (int) hom [d] <<
						" generators of H_" << d <<
						" follow:" << '\n';
			const hashedset<tCell> &cset = g [d];
			for (int i = 0; i < (int) hom [d]; i ++)
			{
				if (format == 0)
					if ((int) hom [d] > 1)
						out << "generator " <<
							(i + 1) << '\n';
				const chain<euclidom> &lst =
					c. gethomgen (d, hom [d]. num (i));
				for (int j = 0; j < (int) lst; j ++)
				{
					const coordinate *left =
						cset [lst. num (j)].
						leftcoord ();
					coordinate right [tCell::MaxDim];
					cset [lst. num (j)].
						rightcoord (right);
					int projdim = 0;
					for (int k = 0; k < xdim; k ++)
						if (left [k] != right [k])
							projdim ++;
					if (projdim != d)
						continue;
					if (format == 0)
						out << lst [j]. coef () <<
							" * " << tCell (left,
							right, xdim) << '\n';
					else
					{
						int k;
						for (k = 0; k < xdim; k ++)
							out << left [k] <<
								" ";
						for (k = 0; k < xdim; k ++)
							out << right [k] <<
								" ";
						out << lst [j]. coef () <<
							" " << (i + 1) <<
							" " << d << '\n';
					}
				}
			}
		}
	return out;
} /* writegenerators */

template <class euclidom, class tCell>
void createprojection (const gcomplex<tCell,euclidom> &Fcompl,
	const gcomplex<tCell,euclidom> &Ycompl,
	chainmap<euclidom> &cmap,
	int offset, int outdim, int discarddim, int *level = NULL)
// Create the chain map of the projection from a cell complex of the graph
// of a map to a cell complex of the codomain of the map.
// If a table of levels is given, create the map only at given levels.
{
	// go through the list of all the dimensions which are of concern
	for (int d = 0; d <= Ycompl. dim (); d ++)
		if ((!level || level [d]) && (Fcompl. dim () >= d))
		{
			// take sets of cells of this dimension
			const hashedset<tCell> &Fset = Fcompl [d];
			if (!(int) Fset)
				continue;
			const hashedset<tCell> &Yset = Ycompl [d];
			if (!(int) Yset)
				continue;

			// go through the list of cells in Fcompl of dim. d
			for (int i = 0; i < (int) Fset; i ++)
			{
				// get this cell and its coordinates
				const tCell &Fcell = Fset [i];
				const coordinate *left = Fcell. leftcoord ();
				coordinate right [tCell::MaxDim];
				Fcell. rightcoord (right);

				// check if this cell has no width in the
				// directions that are discarded
				register int j;
				for (j = 0; j < offset; j ++)
					if (left [j] != right [j])
					{
						j = offset + 33;
						break;
					}
				if (j > offset)
					continue;
				for (j = 0; j < discarddim; j ++)
					if (left [offset + outdim + j] !=
						right [offset + outdim + j])
					{
						j = discarddim + 33;
						break;
					}
				if (j > discarddim)
					continue;

				// create the projected cell
				if (!(PointBase::check (left + offset, outdim)))
					continue;
				if (!(PointBase::check (right + offset, outdim)))
					continue;
				tCell projected (left + offset,
					right + offset, outdim);

				// find its number in Y
				int nr = Yset. getnumber (projected);

				// if not found, discard it
				if (nr < 0)
					continue;

				// add the pair to the projection map
				euclidom e;
				e = 1;
				cmap. add (d, nr, i, e);
			}
		}
	return;
} /* createprojection */

template <class euclidom, class tCell, class tCube>
int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f1, const mvmap<tCube,tCube> &f2,
	const hashedset<tCube> &dom1, const hashedset<tCube> &dom2)
// Create images of cells in 'm' as unions of cubes determined by f1 and f2.
// The domains of interest are given explicitly.
// Return the total number of cubes in all the images of cells.
{
	int countimages = 0;
	coordinate leftbound [tCell::MaxDim];
	coordinate rightbound [tCell::MaxDim];
	coordinate left [tCell::MaxDim];
	coordinate right [tCell::MaxDim];
	for (int d = 0; d <= m. dim (); d ++)
	{
		const hashedset<tCell> &cset = m. get (d);
		if (!(int) cset)
			continue;
		const int spacedim = cset [0]. spacedim ();
		for (int i = 0; i < (int) cset; i ++)
		{
			cset [i]. leftcoord (left);
			cset [i]. rightcoord (right);
			for (int j = 0; j < spacedim; j ++)
			{
				// compensate for space wrapping
				if (right [j] < left [j])
					right [j] = left [j] + 1;
				// compute the bounds
				leftbound [j] = (coordinate) (left [j] -
					(left [j] == right [j]));
				rightbound [j] = (coordinate) (right [j] +
					(left [j] == right [j]));
			}
			rectangle r (leftbound, rightbound, spacedim);
			const coordinate *c;
			while ((c = r. get ()) != NULL)
			{
				if (!PointBase::check (c, spacedim))
					continue;
				tCube q (c, spacedim);
				if (dom1. check (q))
					m. add (d, i, f1 (q));
				if (dom2. check (q))
					m. add (d, i, f2 (q));
			}
			countimages += m (cset [i]). size ();
		}
	}
	return countimages;
} /* createimages */

/// A wrapper for the above function if there is only one map.
template <class euclidom, class tCell, class tCube>
inline int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f, const hashedset<tCube> &dom)
{
	mvmap<tCube,tCube> emptymap;
	hashedset<tCube> emptyset;
	return createimages (m, f, emptymap, dom, emptyset);
} /* createimages */

/// Creates images of cells in m as unions of cubes determined by f1 and f2.
/// Return the total number of cubes in all the images of cells.
template <class euclidom, class tCell, class tCube>
int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f1, const mvmap<tCube,tCube> &f2)
{
	const hashedset<tCube> &dom1 = f1. getdomain ();
	const hashedset<tCube> &dom2 = f2. getdomain ();
	return createimages (m, f1, f2, dom1, dom2);
} /* createimages */

/// A wrapper for the above function if there is only one map.
template <class euclidom, class tCell, class tCube>
inline int createimages (mvcellmap<tCell,euclidom,tCube> &m,
	const mvmap<tCube,tCube> &f)
{
	mvmap<tCube,tCube> emptymap;
	return createimages (m, f, emptymap);
} /* createimages */

/// Creates the image of the projection from the set of cubical cells in the
/// given geometric complex to the subspace of R^n spanned by the 'outdim'
/// subsequent standard vectors with the first number 'offset'.
/// Only these cells which appear in 'only' are added unless 'only' is empty.
template <class euclidom, class tCell>
void project (const gcomplex<tCell,euclidom> &c,
	gcomplex<tCell,euclidom> &img,
	const gcomplex<tCell,euclidom> &only,
	int offset, int outdim, int discarddim, const int *level,
	bool watchforimages)
{
	// go through the list of all the dimensions which are of concern
	for (int d = 0; d <= c. dim (); d ++)
		if ((!level || level [d]) && (c. dim () >= d))
		{
			// take sets of cells of this dimension
			const hashedset<tCell> &cset = c [d];
			if (!(int) cset)
				continue;

			// go through the list of cells in c of dimension d
			for (int i = 0; i < (int) cset; i ++)
			{
				// get this cell and its coordinates
				const tCell &thecell = cset [i];
				const coordinate *left =
					thecell. leftcoord ();
				coordinate right [tCell::MaxDim];
				thecell. rightcoord (right);

				// check if this cell has no width in the
				// directions that are discarded
				int j;
				for (j = 0; j < offset; j ++)
					if (left [j] != right [j])
					{
						j = offset + 33;
						break;
					}
				if (j > offset)
					continue;
				for (j = 0; j < discarddim; j ++)
					if (left [offset + outdim + j] !=
						right [offset + outdim + j])
					{
						j = discarddim + 33;
						break;
					}
				if (j > discarddim)
					continue;

				// add the projected cell to the complex
				if (!(PointBase::check (left + offset, outdim))
					|| !(PointBase::check (right + offset,
					outdim)))
					if (watchforimages)
						throw "Inclusion undefined!";
					else
						continue;
				tCell projected (left + offset,
					right + offset, outdim);
				if ((only. dim () >= 0) &&
					!only. check (projected))
					if (watchforimages)
						throw "Inclusion undefined.";
					else
						continue;
				img. add (projected);
			}
		}
	return;
} /* project */


// --------------------------------------------------
// ------------------- read cubes -------------------
// --------------------------------------------------

std::istream &operator >> (std::istream &in, cubes &c);

std::istream &operator >> (std::istream &in, cubicalmap &m);

// --------------------------------------------------

// read the domain of a multivalued cubical map
std::istream &readdomain (std::istream &in, cubes &dom);

// read the image of a multivalued cubical map
std::istream &readimage (std::istream &in, cubes &img);

// read the image of a set under a multivalued cubical map
std::istream &readimage (std::istream &in, const cubes &dom, cubes &img);

// read the restriction of a multivalued map to the given set(s)
std::istream &readselective (std::istream &in, const cubes &dom,
	cubicalmap &m);
std::istream &readselective (std::istream &in, const cubes &dom1,
	const cubes &dom2, cubicalmap &m);


// --------------------------------------------------
// ------------------- neighbors --------------------
// --------------------------------------------------

// Return the maximal number of neighbors of a cube: 3^dim - 1.
int getmaxneighbors (int dim);

// Return the number of the neighbor bit for the given neighbor of 'q'
// or -1 if not a neighbor or the same cube as 'q'.
int neighbor2bit (const Cube &q, const Cube &neighbor);

// Create the neighbor of the given cube with the specified number.
// If the coordinates of the neighbor do not belong to 'PointBase',
// return 'q', unless 'unconditional'.
Cube bit2neighbor (const Cube &q, int number, bool unconditional = false);

// Indicate the neighbors of the given cube from the given set in the
// bit field and return the number of neighbors.
// If the limit is nonzero, then quit after having found that many neighbors.
int getneighbors (const Cube &q, bitfield *bits, const cubes &set,
	int limit = 0);
int getneighbors (const Cube &q, bitfield *bits, const cubes &set1,
	const cubes &set2, int limit = 0);

// Add neighbors listed in the given bit field to the set of cubes
// unless they are in the 'forbidden' set.
int addneighbors (const Cube &q, const bitfield &bits, cubes &set,
	const cubes &notthese);

// Add neighbors listed in the given bit field to the set of cubes.
// If not 'unconditional', then add only cubes which already exist.
int addneighbors (const Cube &q, const bitfield &bits, cubes &set,
	bool unconditional = false);

// Add intersections of neighbors listed in the given bit field with the
// given cube to the cubical complex.
// If not 'unconditional', then add only cubes which already exist.
int addneighbors (const Cube &q, const bitfield &bits,
	gcomplex<CubicalCell,integer> &c, bool unconditional = false);

/// Uses binary decision diagrams to verify whether the neighborhood
/// of the given cube in the given set is acyclic. Uses the bitfield
/// provided to record each tested neighbor in.
bool bddacyclic (const Cube &q, int dim, const SetOfCubes &s,
	BitField &b);

/// As above - for the union of the two sets.
bool bddacyclic (const Cube &q, int dim, const SetOfCubes &s1,
	const SetOfCubes &s2, BitField &b);

// Check whether this cube's nieghbors form an acyclic set.
// Return: 1 = yes, 0 = probably no, but it is not sure.
int acyclic (const Cube &q, bitfield &b);

// Check thoroughly whether the given set of cubes is acyclic.
// Return true if yes, false if not.
bool acyclic (const cubes &cset);

/// Verifies the acyclicity of a neighborhood of the given cube.
/// \param q - the cube whose neighborhood is verified
/// \param dim - the dimension of the cube
/// \param cset - the set to search for the cubes adjacent to <i>q</i>
/// \param b - a pointer to a bitfield in which neighbors are marked
/// \param maxneighbors - the maximal possible number of neighbors:
/// should be equal to 3^<i>dim</i> - 1
/// Uses BDDs if available. Marks verified neighbors in b.
bool acyclic (const Cube &q, int dim, const SetOfCubes &cset,
	BitField *b, int maxneighbors);

/// The same as above, but checking neighbors in the union of two sets.
bool acyclic (const Cube &q, int dim, const SetOfCubes &cset1,
	const SetOfCubes &cset2, BitField *b, int maxneighbors);


// --------------------------------------------------
// ------------------ reduce cubes ------------------
// --------------------------------------------------

// Leave in the first set only these cubes which are neighbors of at least
// one cube appearing in the second set.
void leaveneighborsonly (cubes &cset, const cubes &nset);

// Verify if the map remains acyclic after the addition or removal of the
// given cube to/from the union of the first and the second set.
// Assume that the map is acyclic before the change.
// Return 'true' if yes for sure, 'false' if there is some doubt about it.
bool remainsacyclic (const cubicalmap &map, const Cube &q,
	const cubes &cset1, const cubes *cset2 = NULL);

// Reduce a pair of sets of cubes for relative homology computation.
// Return the number of cubes removed from both sets.
// Use and update the given bitfields for speeding up computations.
// The sets 'cset' and 'other' must be disjoint and every cube that
// is supposed to be kept in either must appear also in 'keep'.
int cubreducequiet (cubes &cset, cubes &other, const cubes &keep,
	bool quiet = true);
inline int cubreduce (cubes &cset, cubes &other, const cubes &keep)
{
	return cubreducequiet (cset, other, keep, false);
} /* cubreduce */

// The same as above, but make sure that the acyclicity of the given map
// is not spoiled.
int cubreducequiet (cubes &cset, cubes &other, cubicalmap &map,
	const cubes &keep, bool quiet = true);
inline int cubreduce (cubes &cset, cubes &other, cubicalmap &map,
	const cubes &keep)
{
	return cubreducequiet (cset, other, map, keep, false);
} /* cubreduce */

// Reduce the set 'cset' towards 'maincset'. These sets must be disjoint.
// Return the number of cubes removed from 'cset'.
int cubreducequiet (const cubes &maincset, cubes &cset, bool quiet = true);
inline int cubreduce (const cubes &maincset, cubes &cset)
{
	return cubreducequiet (maincset, cset, false);
} /* cubreduce */

// Expand the set 'other' towards 'cset' without changing the homology
// of (cset + other, other). The two sets must be disjoint.
// If a map is given, then increase 'img' if necessary to cover F (other),
// but move cubes only if you can prove that the inclusion of the old 'img'
// into the new, enlarged 'img' induces an isomorphism in homology.
// Every cube added to 'img' is removed from 'imgsrc'.
int cubexpand (cubes &cset, cubes &other, bool quiet = false);
int cubexpand (cubes &cset, cubes &other, cubes &imgsrc, cubes &img,
	const cubicalmap &map, bool indexmap,
	bool checkacyclic, bool quiet = false);

} // namespace homology
} // namespace capd

#endif // _CUBISETS_H_

/// @}
