/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file homtools.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on April 16, 2003. Last revision: May 14, 2006.


#ifndef _CAPD_HOMTOOLS_H_
#define _CAPD_HOMTOOLS_H_

#include "homology/cubisets.h"
#include "homology/simplex.h"
#include "homology/bitmaps.h"

namespace capd {
namespace homology {


// --------------------------------------------------
// ---------------------- READ ----------------------
// --------------------------------------------------

/// Reads all the cubes from the given file and ignores them.
/// In this way the numbers of cubes are established.
inline void scancubes (const char *name)
{
	if (!name || !*name)
		return;
	std::ifstream in (name);
	if (!in)
		fileerror (name);

	// read all the cubes contained in the file,
	// but ignore lines not starting with a parenthesis
	sout << "Scanning the file '" << name << "' for cubes... ";
	int count = 0;
	int dim = -19;
	bool warned = false;
	ignorecomments (in);
	while (in)
	{
		coordinate c [Cube::MaxDim];
		if (closingparenthesis (in. peek ()) == EOF)
			ignoreline (in);
		else
		{
			// read the point from the file
			int d = readcoordinates (in, c, Cube::MaxDim);

			// verify the dimension of the point
			if (dim < 0)
				dim = d;
			else if ((dim != d) && !warned)
			{
				sout << "\nWARNING: Not all the cubes have "
					"the same dimension.\n";
				warned = true;
			}

			// add the point to 'pbase' and verify its number
			int n = pbase::number (c, d);
			if ((n != count) && !warned)
			{
				cube q (c, d);
				sout << "\nWARNING: Some cubes are "
					"repeated - " << q <<
					" for example.\n";
				warned = true;
			}

			// count the point
			count ++;
		}
		ignorecomments (in);
	}
	sout << count << " cubes analyzed.\n";
	return;
} /* scancubes */

// Read a given set from the file and show appropriate messages.
// Assume the elements of the set begin with an opening parenthesis-char
// (or a digit, if "digitOk" == true) and ignore all the preceding data.
// Use the given plural name of the elements for the messages.
// Note: This procedure is suitable for gcomplex<type> or hashedset<type>.
template <class settype>
void readtheset (const char *name, settype &s, const char *pluralname,
	const char *what/*, bool digitOk = false*/)
{
	// if no file name is given, do nothing
	if (!name)
		return;

	// show what you are doing
	sout << "Reading " << pluralname;
	if (what)
		sout << " to " << what;
	sout << " from '" << name << "'... ";

	// open the file
	std::ifstream in (name);
	if (!in)
		fileerror (name);

	// ignore all the introductory data
	ignorecomments (in);
	while (!!in && (closingparenthesis (in. peek ()) == EOF) &&
		((in. peek () < '0') || (in. peek () > '9')) &&
		(in. peek () != '-'))
	{
		ignoreline (in);
		ignorecomments (in);
	}

	// read the set and show how many elements have been read
	int prev = (int) s;
	in >> s;
	sout << ((int) s - prev) << " " << pluralname << " read.\n";

	return;
} /* readtheset */

// Use the general procedure "readtheset" to read a geometric complex.
template <class cell, class euclidom>
inline void readcells (const char *name, gcomplex<cell,euclidom> &s,
	const char *what)
{
	readtheset (name, s, cell::pluralname (), what);
	return;
} /* readcells */

// Use the general procedure "readtheset" to read a set of elements.
template <class element>
inline void readelements (const char *name, hashedset<element> &s,
	const char *what)
{
	readtheset (name, s, element::pluralname (), what);
	return;
} /* readelements */

// Read a set of cubes from the given file (cubes need a special treatment).
inline void readelements (const char *name, cubes &cub, const char *what)
{
	readtheset (name, cub, cube::pluralname (), what);
	return;
} /* readelements */


// --------------------------------------------------
// ---------------------- SAVE ----------------------
// --------------------------------------------------

// Save a given set to a file and show appropriate messages.
// Note: This procedure is suitable for gcomplex<type> or hashedset<type>.
template <class settype>
void savetheset (const char *name, const settype &s, const char *pluralname,
	const char *what, const char *filecomment = 0)
{
	// if no file name is given, do nothing
	if (!name)
		return;

	// show what you are doing
	sout << "Saving " << pluralname;
	if (what)
		sout << " in " << what;
	sout << " to '" << name << "'... ";

	// open the file
	std::ofstream out (name);
	if (!out)
		fileerror (name, "create");
		
	// save the data
	if (filecomment)
		out << filecomment;
	out << s;
	if (!out)
		fileerror (name, "save");

	// show how many elements have been written
	sout << (int) s << " " << pluralname << " written.\n";

	return;
} /* writetheset */

// Use the general procedure "savetheset" to save a geometric complex.
template <class cell, class euclidom>
inline void savecells (const char *name, const gcomplex<cell,euclidom> &s,
	const char *what, const char *filecomment = 0)
{
	savetheset (name, s, cell::pluralname (), what, filecomment);
	return;
} /* savecells */

// Use the general procedure "savetheset" to save a set of elements.
template <class element>
inline void saveelements (const char *name, const hashedset<element> &s,
	const char *what, const char *filecomment = 0)
{
	savetheset (name, s, element::pluralname (), what, filecomment);
	return;
} /* saveelements */

// Save a set of cubes to the given file (cubes need a special treatment).
inline void saveelements (const char *name, const cubes &cub,
	const char *what, const char *filecomment = 0)
{
	savetheset (name, cub, cube::pluralname (), what, filecomment);
	return;
} /* saveelements */


// --------------------------------------------------
// --------------------- VERIFY ---------------------
// --------------------------------------------------

// Check if X is a subset of the union of Y and B. Display messages
// and a warning if necessary. Return the result: true = passed.
bool checkinclusion (const cubes &Xcubes, const cubes &Ycubes,
	const cubes &Bcubes, const char *Xname, const char *YBname);

// The same as above, but check for the inclusion of X in Y only.
inline bool checkinclusion (const cubes &Xcubes, const cubes &Ycubes,
	const char *Xname, const char *Yname)
{
	cubes empty;
	return checkinclusion (Xcubes, Ycubes, empty, Xname, Yname);
} /* checkinclusion */

// Check if the image of X by F is contained in the union of Y and B.
// Display messages and a warning if necessary. Return true if yes.
bool checkimagecontained (const cubicalmap &Fcubmap, const cubes &Xcubes,
	const cubes &Ycubes, const cubes &Bcubes,
	const char *Xname, const char *Yname);

// Check if the image of X by F is contained in the set Y alone.
// Display messages and a warning if necessary. Return true if yes.
inline bool checkimagecontained (const cubicalmap &Fcubmap,
	const cubes &Xcubes, const cubes &Ycubes,
	const char *Xname, const char *Yname)
{
	cubes empty;
	return checkimagecontained (Fcubmap, Xcubes, Ycubes, empty,
		Xname, Yname);
} /* checkimagecontained */

// Check if the image of A by F is disjoint from Y (actually, from Y\\B).
// Display messages and a warning if necessary. Return true if yes.
bool checkimagedisjoint (const cubicalmap &Fcubmap, const cubes &Acubes,
	const cubes &Ycubes, const char *Aname, const char *Yname);

// Check if the image of each element of the domain of this map is acyclic.
// Show messages and a warning if necessary. Return true = yes.
bool checkacyclicmap (const mvcellmap<qcell,integer,cube> &Fcellcubmap,
	const char *Xname);


// --------------------------------------------------
// --------------------- REDUCE ---------------------
// --------------------------------------------------

// Restrict the set of cubes 'Acubes' to these cubes which are neighbors
// of any of the cubes in 'Xcubes' and display appropriate messages.
void restrictAtoneighbors (cubes &Xcubes, cubes &Acubes,
	const char *Xname, const char *Aname, const cubes *keepcubes = NULL);

// Remove 'Acubes' from 'Xcubes'.
void removeAfromX (cubes &Xcubes, const cubes &Acubes,
	const char *Xname, const char *Aname);

// Remove from 'X' all the cells that appear in 'A'.
template <class cell, class euclidom>
void removeAfromX (gcomplex<cell,euclidom> &X,
	const gcomplex<cell,euclidom> &A,
	const char *Xname, const char *Aname)
{
	if (!(int) X || !(int) A)
		return;
	sout << "Computing " << Xname << "\\" << Aname << "... ";
	int prev = (int) X;
	X. remove (A);
	sout << (prev - (int) X) << ' ' << cell::pluralname () <<
		" removed from " << Xname << ", " << (int) X << " left.\n";
	return;
} /* removeAfromX */

// Expand the other element of the pair into the main portion of the set.
void expandAinX (cubes &Xcubes, cubes &Acubes,
	const char *Xname, const char *Aname);

// Expand the other element of the pair into the main portion of the set.
void expandAinX (cubes &Xcubes, cubes &Acubes, cubes &Ycubes,
	cubes &Bcubes, const cubicalmap &Fcubmap,
	const char *Xname, const char *Aname, const char *Bname,
	bool indexmap, bool checkacyclic);

// Reduce the pair of sets of cubes. Keep the given cubes untouched.
void reducepair (cubes &Xcubes, cubes &Acubes, const cubes &Xkeepcubes,
	const char *Xname, const char *Aname);

// Reduce the pair of sets of cubes. Keep the given cubes untouched.
// Make sure that the acyclicity of the given map is not spoiled.
void reducepair (cubes &Xcubes, cubes &Acubes, cubicalmap &Fcubmap,
	const cubes &Xkeepcubes, const char *Xname, const char *Aname);

// Add the images of both maps to the set of cubes to be kept.
// Include the sets 'Xcubes' and 'Acubes' if this is an index map.
void addmapimg (const cubicalmap &Fcubmap, const cubicalmap &FcubmapA,
	const cubes &Xcubes, const cubes &Acubes, cubes &Ykeepcubes,
	bool indexmap);

// Transform cubes to cells.
void cubes2cells (cubes &Xcubes, cubicalcomplex &Xcompl,
	const char *Xname, bool deletecubes = true);

// Collapse a pair of geometric complexes.
template <class cell, class euclidom>
void collapse (gcomplex<cell,euclidom> &Xcompl,
	gcomplex<cell,euclidom> &Acompl, gcomplex<cell,euclidom> &Xkeepcompl,
	const char *Xname, const char *Aname, bool addbd = true,
	bool addcob = false, bool disjoint = true, const int *level = NULL)
{
	// say what you are about to do
	sout << "Collapsing faces in " << Xname;
	if ((int) Acompl)
		sout << " and " << Aname;
	sout << "... ";

	// collapse the faces as requested to
	int count = Xcompl. collapse (Acompl, Xkeepcompl,
		addbd, addcob, disjoint, level);

	// say something about the result
	sout << 2 * count << " removed, " << (int) Xcompl << " left.\n";

	// if something remains in A, say it
	if ((int) Acompl)
		sout << "There are " << (int) Acompl << " faces "
			"of dimension up to " << Acompl. dim () <<
			" left in " << Aname << ".\n";

	return;
} /* collapse */

// Collapse a pair of geometric complexes if no complex to keep is requested.
template <class cell, class euclidom>
inline void collapse (gcomplex<cell,euclidom> &Xcompl,
	gcomplex<cell,euclidom> &Acompl,
	const char *Xname, const char *Aname, bool addbd = true,
	bool addcob = false, bool disjoint = true, const int *level = NULL)
{
	gcomplex<cell,euclidom> empty;
	collapse (Xcompl, Acompl, empty, Xname, Aname,
		addbd, addcob, disjoint, level);
	return;
} /* collapse */

// Collapse a single geometric complex.
template <class cell, class euclidom>
inline void collapse (gcomplex<cell,euclidom> &Xcompl,
	gcomplex<cell,euclidom> &Xkeepcompl,
	const char *Xname, bool addbd = true, bool addcob = false,
	bool disjoint = true, const int *level = NULL)
{
	gcomplex<cell,euclidom> empty;
	collapse (Xcompl, empty, Xkeepcompl, Xname, "",
		addbd, addcob, disjoint, level);
	return;
} /* collapse */

// Collapse a single geometric complex if no complex to keep is requested.
template <class cell, class euclidom>
inline void collapse (gcomplex<cell,euclidom> &Xcompl,
	const char *Xname, bool addbd = true, bool addcob = false,
	bool disjoint = true, const int *level = NULL)
{
	gcomplex<cell,euclidom> empty;
	collapse (Xcompl, empty, empty, Xname, "",
		addbd, addcob, disjoint, level);
	return;
} /* collapse */

// Decrease the dimension of the geometric complex by adding boundary cells
// to all the cells on higher dimensions and then removing these cells.
template <class cell, class euclidom>
void decreasedimension (gcomplex<cell,euclidom> &Acompl,
	int dim, const char *name)
{
	if (Acompl. dim () <= dim)
		return;
	if (dim < 0)
		dim = 0;
	sout << "Adding to " << name << " boundaries of high-dim " <<
		cell::pluralname () << "... ";
	int howmany = 0;
	for (int i = Acompl. dim (); i > dim; i --)
	{
		sout << '.';
		howmany += Acompl. addboundaries (i);
		Acompl. removeall (i);
	}
	sout << ' ' << howmany << " added.\n";
	return;
} /* decreasedimension */

// Add boundaries to the geometric complex X or to both X and A.
template <class cell, class euclidom>
void addboundaries (gcomplex<cell,euclidom> &Xcompl,
	gcomplex<cell,euclidom> &Acompl,
	int minlevel, bool bothsets, const char *Xname, const char *Aname)
{
	// if there are no cells in the complex, there is nothing to do
	if (!(int) Xcompl)
		return;

	// say what you are doing
	sout << "Adding boundaries of " << cell::pluralname () << " in ";
	if ((int) Acompl)
		sout << Xname << " and " << Aname << "... ";
	else
		sout << Xname << "... ";

	// add the boundaries and count the number of added cells
	int howmany = 0;
	for (int i = Xcompl. dim (); (i >= minlevel) && i; i --)
	{
		if (bothsets)
		{
			howmany += Xcompl. addboundaries (i, true);
			if (Acompl. dim () >= i)
				howmany += Acompl. addboundaries (i,
					true);
		}
		else
			howmany += Xcompl. addboundaries (i, Acompl);
	}

	// show the total number of added cells
	sout << howmany << ' ' << cell::pluralname () << " added.\n";
	return;
} /* addboundaries */

// Add boundaries to X or to X and A.
//void addboundaries (cubicalcomplex &Xcompl, cubicalcomplex &Acompl,
//	int minlevel, bool bothsets, const char *Xname, const char *Aname);

// Add boundaries to X or to X and A.
//void addboundaries (simplicialcomplex &Xcompl, simplicialcomplex &Acompl,
//	int minlevel, bool bothsets, const char *Xname, const char *Aname);


// --------------------------------------------------
// ---------------- READ BITMAP FILE ----------------
// --------------------------------------------------

// Read a set of squares from a bitmap file to a set of cubes.
// Each pixel whose RGB value is within the given range
// gives raise to one square with the corresponding coordinates.
// Return 0 on success, throw an error message if failed.
int ReadBitmapFile (const char *bmpname, SetOfCubes &cset,
	long mincolor = 0, long maxcolor = 0x7F7F7FL);

} // namespace homology
} // namespace capd

#endif // _HOMTOOLS_H

/// @}
