/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file pointbas.h
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: November 14, 2004.


#ifndef _CAPD_POINTBAS_H_
#define _CAPD_POINTBAS_H_

#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"

#include <iostream>
#include <fstream>
#include <cstdlib>

namespace capd {
namespace homology {


// classes defined within this header file (in this order):
template <class coordtype>
class tPointBase;

template <class coordtype>
class tPointBaseInitializer;

// default types of templates
typedef tPointBase<coordinate> PointBase;

// some abbreviations or lower-case versions of the data type names
typedef PointBase pbase;

// define the number of signed bits to store the dimension (i.e., 6: max 31);
// the maximal allowed dimension in the program is less than this
const int DimBits = 6;

// based on the above, define the number of remaining bits and their mask
const int NumBits = ((int) (sizeof (int) << 3) - DimBits);
const int NumMask = (0xFFFFFFFFu >> DimBits);

// based on the above, define the maximal allowed dimension
const int MaxDim1 = (1 << (DimBits - 1));
const int MaxDim2 = (32 - DimBits);
const int MaxBasDim = (MaxDim1 < MaxDim2) ? MaxDim1 : MaxDim2;


// --------------------------------------------------
// ------------------- PointBase --------------------
// --------------------------------------------------
// Set of points which is a base for other classes.
// This base is used to give each point its unique number.

template <class coordtype>
class tPointBase
{
public:
	// get the number of the point with given coordinates;
	// points for each dimension get numbers starting at 1
	static int number (const coordtype *c, int d);

	// check whether this point already exists;
	// return 0 if not, 1 if (possibly) yes;
	// do not add the point to the structure
	static int check (const coordtype *c, int d);

	// retrieve the coordinates of the given point
	static const coordtype *coord (int nr, int d);

	// determine the default dimension for points
	static int defaultdimension (void);

	// set the pointset to be quiet or not
	static void quiet (bool what = true);
	
	// set space wrapping for all the dimensions beginning
	// with 'mindim' and ending at 'maxdim - 1'
	static void wrap (const coordtype *c,
		int mindim = 1, int maxdim = MaxBasDim);
	static void wrap (coordtype c,
		int mindim = 1, int maxdim = MaxBasDim);
	// get space wrapping for the given dimension
	static const coordtype *wrap (int d);

	// wrap the given coordinates if necessary
	static void wrapcoord (coordtype *c, int dim);

	// copy coordinates and wrap them if necessary
	static void wrapcopy (coordtype *dest, const coordtype *src,
		int dim);

	// reset the data; note: cubes and cells created before
	// resetting are unusable after the reset
	static void reset (void);

	// forget base points, free memory and make cubes unusable
	static void forget (void);

	// show vertices used
	static outputstream &showused (outputstream &out);
	static std::ostream &showused (std::ostream &out);

	// friendship declarations
	friend class tPointBaseInitializer<coordtype>;

	// are there any cubes other than PointBased in use?
	static bool outOfControl;

protected:
	// the sets of points from which numbers are taken
	static tPointset<coordtype> **p;

	// the number of allocated sets of points
	static int n;

	// should the summary of points' usage be displayed?
	static bool show;

	// were the base points forgot?
	static bool forgotten;

	// an object responsible for deleting the pointsets at exit
	static tPointBaseInitializer<coordtype> pointBaseInitializer;

}; /* class tPointBase */

// --------------------------------------------------

template <class coordtype>
inline const coordtype *tPointBase<coordtype>::coord (int nr, int d)
{
	if (forgotten)
		throw "Trying to retrieve forgotten coordinates.";
	if ((d > n) || (d <= 0) || (p [d - 1] == NULL))
		return NULL;
	else
		return (*p [d - 1]) [nr];
} /* tPointBase::coord */

template <class coordtype>
inline int tPointBase<coordtype>::check (const coordtype *c, int d)
{
	if (outOfControl)
		return 1;
	if (forgotten)
		throw "Trying to check forgotten coordinates.";
	if ((d > n) || (d <= 0) || (p [d - 1] == NULL))
		return 0;
	return p [d - 1] -> check (c);
} /* tPointBase::check */

template <class coordtype>
inline void tPointBase<coordtype>::quiet (bool what)
{
	show = !what;
	return;
} /* tPointBase::quiet */

template <class coordtype>
inline int tPointBase<coordtype>::number (const coordtype *c, int d)
{
	if (forgotten)
		throw "Trying to find the number of forgotten coordinates.";

	if (d < 0)
		return -1;
	if (d >= MaxBasDim)
		throw "Dimension too high.";

	// enhance the table of sets of points if necessary
	if (d > n)
	{
		tPointset<coordtype> **newtable =
			new tPointset<coordtype> * [d];
		for (int i = 0; i < d; i ++)
			newtable [i] = (i < n) ? p [i] : NULL;
		if (p)
			delete [] p;
		p = newtable;
		n = d;
	}

	if (!p [d - 1])
	{
		p [d - 1] = new tPointset<coordtype> (1024);
		p [d - 1] -> dimension (d);
	}

	int number = p [d - 1] -> add (c);
	if (number > NumMask)
		throw "Too many points.";

	return number;
} /* tPointBase::number */

template <class coordtype>
inline void tPointBase<coordtype>::wrap (const coordtype *c, int mindim,
	int maxdim)
{
	if (forgotten)
		throw "Trying to wrap forgotten coordinates.";

	// correct the left and right bound for dimensions
	if (mindim < 1)
		mindim = 1;
	if (maxdim > MaxBasDim)
		maxdim = MaxBasDim;

	// enhance the table of sets of points if necessary
	if (maxdim > n)
	{
		tPointset<coordtype> **newtable =
			new tPointset<coordtype> * [maxdim];
		for (int i = 0; i < maxdim; i ++)
			newtable [i] = (i < n) ? p [i] : NULL;
		if (p)
			delete [] p;
		p = newtable;
		n = maxdim;
	}

	// set wrapping coordinates and allocate sets of points if needed
	for (int d = mindim; d < maxdim; d ++)
	{
		if (!p [d - 1])
		{
			p [d - 1] = new tPointset<coordtype>;
			p [d - 1] -> dimension (d);
		}
		p [d - 1] -> wrapspace (c);
	}

	return;
} /* tPointBase::wrap */

template <class coordtype>
inline void tPointBase<coordtype>::wrap (coordtype c, int mindim, int maxdim)
{
	coordtype wraptable [MaxBasDim];
	for (int i = 0; i < MaxBasDim; i ++)
		wraptable [i] = c;
	wrap (wraptable, mindim, maxdim);
	return;
} /* tPointBase::wrap */

template <class coordtype>
inline const coordtype *tPointBase<coordtype>::wrap (int d)
{
	if (forgotten)
		throw "Trying to get wrapping of forgotten coordinates.";

	if ((d <= 0) || (d - 1 >= n) || !p [d - 1])
		return NULL;
	else
		return p [d - 1] -> wrapspace ();
} /* tPointBase::wrap */

template <class coordtype>
inline void tPointBase<coordtype>::wrapcoord (coordtype *c, int dim)
{
	if ((dim > n) || (dim <= 0) || (p [dim - 1] == NULL))
		return;
	const coordtype *cw = p [dim - 1] -> wrapspace ();
	if (cw)
		capd::homology::wrapcoord (c, c, cw, dim);
	return;
} /* tPointBase::wrapcoord */

template <class coordtype>
inline void tPointBase<coordtype>::wrapcopy (coordtype *dest,
	const coordtype *src, int dim)
{
	const coordtype *cw;
	if ((dim > n) || (dim <= 0) || (p [dim - 1] == 0))
		cw = 0;
	else
		cw = p [dim - 1] -> wrapspace ();
	if (cw)
		capd::homology::wrapcoord (dest, src, cw, dim);
	else
		copycoord (dest, src, dim);
	return;
} /* tPointBase::wrapcopy */

template <class coordtype>
inline void tPointBase<coordtype>::reset (void)
{
	for (int i = 0; i < n; i ++)
		if (p [i])
		{
			delete p [i];
			p [i] = NULL;
		}
	delete [] p;
	p = NULL;
	n = 0;
	forgotten = false;
	return;
} /* tPointBase::reset */

template <class coordtype>
inline void tPointBase<coordtype>::forget (void)
{
	if (forgotten)
		throw "Trying to forget already forgotten coordinates.";

	reset ();
	forgotten = true;
	return;
} /* tPointBase::forget */

// --------------------------------------------------

template <class coordtype>
tPointset<coordtype> **tPointBase<coordtype>::p = NULL;

template <class coordtype>
int tPointBase<coordtype>::n = 0;

template <class coordtype>
bool tPointBase<coordtype>::show = true;

template <class coordtype>
bool tPointBase<coordtype>::forgotten = false;

template <class coordtype>
bool tPointBase<coordtype>::outOfControl = false;

template <class coordtype>
int tPointBase<coordtype>::defaultdimension (void)
{
	static int dim = 0;
	if (dim)
		return dim;
	for (int d = 1; d <= n; d ++)
	{
		if (p [d - 1] == NULL)
			continue;
		if (!(int) (*(p [d - 1])))
			continue;
		return d;
	}
	return 0;
} /* tPointBase::defaultdimension */

template <class coordtype>
outputstream &tPointBase<coordtype>::showused (outputstream &out)
{
	if (forgotten)
		throw "Trying to show forgotten coordinates.";

	bool shown = false;
	for (int i = 0; i < n; i ++)
	{
		if (!p [i])
			continue;
		const tPointset<coordtype> &pset = *(p [i]);
		if (!(int) pset)
			continue;
		out << (shown ? ", " : "Vertices used: ") <<
			(int) pset << " of dim " << (i + 1);
		shown = true;
	}
	if (shown)
		out << ".\n";
	return out;
} /* tPointBase::showused */

template <class coordtype>
std::ostream &tPointBase<coordtype>::showused (std::ostream &out)
{
	outputstream tout (out);
	showused (tout);
	return out;
} /* tPointBase::showused */

// --------------------------------------------------

template <class coordtype>
class tPointBaseInitializer
{
public:
	// deallocate memory kept in the static variables
	// of the class "tPointBase"
	~tPointBaseInitializer ();

}; /* PointBaseInitializer */

template <class coordtype>
tPointBaseInitializer<coordtype> tPointBase<coordtype>::pointBaseInitializer;

template <class coordtype>
tPointBaseInitializer<coordtype>::~tPointBaseInitializer<coordtype> ()
{
	if (tPointBase<coordtype>::forgotten)
		return;
	if (tPointBase<coordtype>::show && tPointBase<coordtype>::n)
		tPointBase<coordtype>::showused (sout);
	tPointBase<coordtype>::forget ();
	return;
} /* PointBaseInitializer::~PointBaseInitializer */

} // namespace homology
} // namespace capd


#endif // _POINTBAS_H_

/// @}
