/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file cubisets.cpp
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started in January 2002. Last revision: October 13, 2006.


#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/pointset.h"
#include "homology/chains.h"
#include "homology/bitfield.h"
#include "homology/integer.h"
#include "homology/hashsets.h"
#include "homology/gcomplex.h"
#include "homology/cubisets.h"

#include <iostream>
#include <fstream>
#include <iomanip>
#include <ctime>
#include <cstdlib>
#include <cctype>

namespace capd {
namespace homology {


// --------------------------------------------------
// ---------------- count neighbors -----------------
// --------------------------------------------------

int getmaxneighbors (int dim)
{
	static int dimension = 0;
	static int maxn = 0;

	if (dim == dimension)
		return maxn;

	maxn = 1;
	for (int i = 0; i < dim; i ++)
		maxn *= 3;
	maxn --;

	dimension = dim;

	return maxn;
} /* getmaxneighbors */


// --------------------------------------------------
// ------------------- BitFields --------------------
// --------------------------------------------------

BitFields::~BitFields ()
{
	for (int i = 0; i < len; i ++)
	{
		if (tab [i])
			delete tab [i];
	}
	if (len)
	{
		delete [] tab;
		delete [] maxkb;
	}
	return;
} /* BitFields::~BitFields */

void BitFields::extend (int n)
{
	// if there is no need to extend the tables, do nothing
	if (n <= len)
		return;

	// allocate new tables
	SetOfBitFields **newtab = new SetOfBitFields * [n];
	int *newmaxkb = new int [n];
	if (!newtab || !newmaxkb)
		throw "Can't extend the table of bitfields.";

	// copy the old tables
	int i;
	for (i = 0; i < len; i ++)
	{
		newtab [i] = tab [i];
		newmaxkb [i] = maxkb [i];
	}

	// initialize the new entries of the tables
	for (i = len; i < n; i ++)
	{
		newtab [i] = NULL;
		newmaxkb [i] = maxkb ? maxkb [0] : -1;
	}

	// deallocate the previous tables
	if (len)
	{
		delete [] tab;
		delete [] maxkb;
	}

	// assign the new values to the data structures
	len = n;
	tab = newtab;
	maxkb = newmaxkb;
	return;
} /* BitFields::extend */

void BitFields::setkblimit (int limit, int dim)
{
	// if no limit is given, do nothing
	if (limit < 0)
		return;

	// extend the tables if necessary
	if (dim >= len)
		extend (dim + 1);

	// set the default limit for every dimension if requested to
	if (dim <= 0)
	{
		if (!len)
			extend (1);
		for (int i = 0; i < len; i ++)
			if ((dim < 0) || (maxkb [i] < 0))
				maxkb [i] = limit;
	}
	// otherwise, set the limit for the given dimension
	else
		maxkb [dim] = limit;

	return;
} /* BitFields::setkblimit */

int BitFields::getkblimit (int dim)
{
	// if no bit field sets are defined, then there is no limit
	if (!len)
		return -1;

	// if the dimension is too high, return the default limit
	else if (dim >= len)
		return maxkb [0];

	// if the dimension given is negative, return default limit
	else if (dim < 0)
		return maxkb [0];

	// otherwise, return the limit for the given dimension
	else
		return maxkb [dim];
} /* BitFields::getkblimit */

void BitFields::forget (int dim)
{
	// if the dimension is out of range, there is nothing to be forgotten
	if (dim >= len)
		return;

	// if the dimension is negative, forget everything
	if (dim < 0)
	{
		for (int i = 0; i < len; i ++)
			forget (i);
		return;
	}

	// if there is not bit field set, do nothing
	if (!tab [dim])
		return;

	// display the bit field usage
	if (tab [dim] -> used ())
		sout << tab [dim] -> used () <<
			" bit fields for dimension " << dim <<
			" were used.\n";

	// forget the given table
	delete tab [dim];
	tab [dim] = NULL;

	return;
} /* BitFields::forget */

void BitFields::allocate (int dim)
{
	// if the dimension is out of range, do nothing
	if (dim < 0)
		return;

	// extend the tables if necessary
	extend (dim + 1);

	// determine the number of bits to be allocated in each bit field
	int maxXneighbors = 1;
	for (int i = 0; i < dim; i ++)
		maxXneighbors *= 3;
	maxXneighbors --;

	// determine a recommended number of bit fields to be allocated
	int recommended [] = {0, 10, 300, 50000, 200000};
	int nXbits = (dim < 5) ? recommended [dim] :
		(((dim + dim / 2) * 1024 * 1024) /
		((maxXneighbors + 7) / 8));

	// adjust the number of bit fields to fit the memory limit
	int memorylimit = maxkb [dim];
	if (memorylimit >= 0)
		nXbits = 1024 * memorylimit /
			((maxXneighbors + 7) / 8);

	// delete the previously allocated set of bit fields if any
	if (tab [dim])
		forget (dim);

	// allocate the set of bit fields
	tab [dim] = nXbits ?
		new SetOfBitFields (maxXneighbors, nXbits) : NULL;

	// show a message of what has been done
	if (nXbits)
		sout << nXbits << " bit fields allocated (" <<
			((((maxXneighbors + 7) / 8) * nXbits + 512 * 1024) /
			1024 / 1024) << " MB) to speed up " << dim <<
			"-dimensional reduction.\n";
	return;
} /* BitFields::allocate */

// the global table of BitFields which store the acyclicity information
BitFields knownbits;

// a special object which will deallocate any left bit field sets
// and display statistics (if necessary) at the program's termination
/*
class KnownBitsDeallocator
{
public:
	KnownBitsDeallocator () {return;};
	~KnownBitsDeallocator () {knownbits. forget (); return;}

} theKnownBitsDeallocator;
*/


// --------------------------------------------------
// ------------ tabulated neighborhoods -------------
// --------------------------------------------------

char *tabulated1d = "\x06";

char *tabulated2d =
	"\x06\x03\x05\x00\x0f\x0a\x0c\x08\x03\x02\x00\x00\x0f\x0a\x0c\x08"
	"\x05\x00\x04\x00\x0f\x0a\x0c\x08\x00\x00\x00\x00\x0f\x0a\x0c\x08";

Tabulated::Tabulated ()
{
	int three = 1;
	for (int dim = 0; dim < maxdim; ++ dim)
	{
		if (dim == 0)
			size [dim] = 0;
		else if (dim == 1)
			size [dim] = 1;
		else
			size [dim] = 1 << (three - 1 - 3);
		deallocate [dim] = false;
		tables [dim] = 0;
	//	sout << "DEBUG: Dim " << dim << ": " << size [dim] <<
	//		" byte(s), " << (three - 1) << " neighbors.\n";
		three *= 3;
	}
	define (1, tabulated1d);
	define (2, tabulated2d);
	return;
} /* Tabulated::Tabulated */

Tabulated::~Tabulated ()
{
	for (int i = 0; i < maxdim; ++ i)
		if (tables [i] && deallocate [i])
			delete [] (tables [i]);
	return;
} /* Tabulated::Tabulated */

int Tabulated::read (int dim, const char *filename)
{
	if ((dim <= 0) || (dim >= maxdim))
		return -1;
	if (!tables [dim] || !deallocate [dim])
	{
		tables [dim] = new char [size [dim]];
		deallocate [dim] = true;
	}
	std::ifstream f;
	f. open (filename, std::ios::in | std::ios::binary);
	f. read (tables [dim], size [dim]);
	if (!f)
		return -1;
	return 0;
} /* Tabulated::read */

int Tabulated::write (int dim, const char *filename) const
{
	if ((dim <= 0) || (dim >= maxdim) || (!tables [dim]))
		return -1;
	std::ofstream f;
	f. open (filename, std::ios::out | std::ios::binary);
	f. write (tables [dim], size [dim]);
	if (!f)
		return -1;
	return 0;
} /* Tabulated::write */

// --------------------------------------------------

Tabulated tabulated;


// --------------------------------------------------
// ------------- neighborhoods for BDDs -------------
// --------------------------------------------------

template <class cubetype, class settype>
class Neighbors
{
public:
	// the default constructor
	Neighbors (const cubetype &middle, const settype &cset,
		BitField &bfield): q (middle), s (cset), b (bfield) {};

	// the procedure for checking whether the given neighbor exists;
	// the number of the neighbor is consistent with the "bit2neighbor"
	// and "neighbor2bit" procedures
	bool check (int n) const
	{
		cubetype neighbor = bit2neighbor (q, n);
		if (neighbor == q)
			return false;
		b. set (n);
		return (s. check (neighbor));
	}

private:
	// the cube whose neighbors are verified
	const cubetype &q;

	// the set of cubes in which the neighbors of the cube are sought
	const settype &s;

	// the bitfield to record each neighbor in
	BitField &b;

}; /* class Neighbors */

template <class cubetype, class settype>
class Neighbors2
{
public:
	// the default constructor
	Neighbors2 (const cubetype &middle, const settype &cset1,
		const settype &cset2, BitField &bfield):
		q (middle), s1 (cset1), s2 (cset2), b (bfield) {};

	// the procedure for checking whether the given neighbor exists;
	// the number of the neighbor is consistent with the "bit2neighbor"
	// and "neighbor2bit" procedures
	bool check (int n) const
	{
		cubetype neighbor = bit2neighbor (q, n);
		if (neighbor == q)
			return false;
		else return (s1. check (neighbor) || s2. check (neighbor));
	}

private:
	// the cube whose neighbors are verified
	const cubetype &q;

	// the sets of cubes in which the neighbors of the cube are sought
	const settype &s1;
	const settype &s2;

	// the bitfield to record each neighbor in
	BitField &b;

}; /* class Neighbors2 */

// --------------------------------------------------

/// The maximal dimension for which the binary decision diagrams are used.
int MaxBddDimPossible = 3;
int MaxBddDim = MaxBddDimPossible;

bool UseMalandainCode = false;

// --------------------------------------------------
// -------------------- acyclic ---------------------
// --------------------------------------------------

/// Verify whether the neighborhood of the given cube is acyclic or not
/// using the code generated from binary decision diagrams.
bool bddacyclic (const Cube &q, int dim, const SetOfCubes &s,
	BitField &b)
{
	switch (dim)
	{
	case 1:
	{
		Neighbors<Cube,SetOfCubes> n (q, s, b);
		return acyclic1d (n);
	//	break;
	}
	case 2:
	{
		Neighbors<Cube,SetOfCubes> n (q, s, b);
		return acyclic2d (n);
	//	break;
	}
	case 3:
	{
		Neighbors<Cube,SetOfCubes> n (q, s, b);
		if (UseMalandainCode)
			return acyclic3d_Malandain (n);
		else
			return acyclic3d (n);
	//	break;
	}
	}
	throw "Trying to use a BDD for dimension other than 1-3.";
} /* bddacyclic */

// As above - a version for the union of two sets.
bool bddacyclic (const Cube &q, int dim, const SetOfCubes &s1,
	const SetOfCubes &s2, BitField &b)
{
	switch (dim)
	{
	case 1:
	{
		Neighbors2<Cube,SetOfCubes> n (q, s1, s2, b);
		return acyclic1d (n);
	//	break;
	}
	case 2:
	{
		Neighbors2<Cube,SetOfCubes> n (q, s1, s2, b);
		return acyclic2d (n);
	//	break;
	}
	case 3:
	{
		Neighbors2<Cube,SetOfCubes> n (q, s1, s2, b);
		if (UseMalandainCode)
			return acyclic3d_Malandain (n);
		else
			return acyclic3d (n);
	//	break;
	}
	}
	throw "Trying to use a BDD for dimension other than 1-3.";
} /* bddacyclic */


// --------------------------------------------------
// ------------------- read cubes -------------------
// --------------------------------------------------

std::istream &operator >> (std::istream &in, SetOfCubes &s)
// An alternative version for reading a set of cubes.
// The only difference is that the word "dimension" is ignored.
{
	// ignore any comments at the beginning of the file
	ignorecomments (in);

	// if the word "dimension" found, ignore the entire line
	if (in. peek () == 'd')
	{
		in. ignore (20000, '\n');
		ignorecomments (in);
	}

	return read (in, s, LARGE_SIZE);
} /* operator >> */

std::istream &operator >> (std::istream &in, cubicalmap &m)
{
	// process the entire input file and read the map line-by-line
	ignorecomments (in);
	while (in. peek () != EOF)
	{
		// ignore all the lines which do not define a map assignment
		while ((closingparenthesis (in. peek ()) == EOF) &&
			((in. peek () < '0') || (in. peek () > '9')) &&
			(in. peek () != EOF))
		{
			ignoreline (in);
			ignorecomments (in);
		}

		// if the end of the file has been reached, exit the loop
		if (in. peek () == EOF)
			break;
	
		// determine the closing parenthesis corresp. to this one
		int closing = closingparenthesis (in. peek ());

		// if the opening parenthesis is valid, read it and
		// check the next character to determine the assignment type
		if (closing != EOF)
		{
			in. get ();
			ignorecomments (in);
		}

		// if the assignment is in the general form, decode the line
		if ((closing == EOF) ||
			(closingparenthesis (in. peek ()) == EOF))
		{
			// read the domain element
			Cube e;
			// if it is given as a number, read it directly
			if (closing == EOF)
			{
				in >> e;
				if (!in)
					throw "Can't read cube's number.";
			}
			// otherwise read the coordinates of the cube
			else
			{
				coordinate coord [Cube::MaxDim];
				int dim = readcoordinates (in, coord,
					Cube::MaxDim, closing);
				if (!in || (dim <= 0))
					throw "Unable to read a cube.";
				e = Cube (coord, dim);
			}
			ignorecomments (in);

			// read the assignment arrow
			while (in. peek () == '-')
				in. get ();
			if (in. peek () == '>')
				in. get ();
			ignorecomments (in);

			// read the image of the cube
			read (in, m [e], SMALL_SIZE);
			ignorecomments (in);
		}

		// otherwise read the assignment in the Jacek & Marcin format
		else
		{
			// read the argument cell
			coordinate argleft [Cube::MaxDim];
			coordinate argright [Cube::MaxDim];
			int dim = readcoordinates (in, argleft, Cube::MaxDim);
			ignorecomments (in);
			int d1 = readcoordinates (in, argright, Cube::MaxDim);
			ignorecomments (in);

			// read the closing and opening brackets
			in. get ();
			ignorecomments (in);
			in. get ();
			ignorecomments (in);

			// read the value cell
			coordinate vleft [Cube::MaxDim];
			coordinate vright [Cube::MaxDim];
			int d2 = readcoordinates (in, vleft, Cube::MaxDim);
			ignorecomments (in);
			int d3 = readcoordinates (in, vright, Cube::MaxDim);
			ignorecomments (in);

			// if there was an I/O error, interrupt reading here
			if (!in || (in. peek () == EOF))
				throw "Cannot read a map assignment line.";

			// read the closing bracket
			in. get ();
			ignorecomments (in);

			// check that all the dimensions are the same
			if ((d1 != dim) || (d2 != dim) || (d3 != dim))
				throw "Wrong dimensions of vertices.";

			// verify that the argument cube is of the right size
			for (int i = 0; i < dim; i ++)
				if (argright [i] - argleft [i] != 1)
					throw "Wrong size of an argument.";

			// add the argument cube to the map's domain
			hashedset<Cube> &v = m [Cube (argleft, dim)];

			// form a rectangle from this value cell
			rectangle r (vleft, vright, dim);

			// add all the value cubes to the image of this element
			const coordinate *c;
			while ((c = r. get ()) != NULL)
				v. add (Cube (c, dim));
		}
	}
	
	return in;
} /* operator >> */

// --------------------------------------------------

std::istream &readdomain (std::istream &in, SetOfCubes &dom)
// Read the domain of a multivalued cubical map.
{
	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		cubicalmap Fdummy;
		return readdomain (in, dom, Fdummy);
	}

	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			Cube q;
			in >> q;
			if (!in)
				throw "Can't read the file.";
			dom. add (q);
		}
		ignoreline (in);
		ignorecomments (in);
	}
	
	return in;
} /* readdomain */

std::istream &readimage (std::istream &in, SetOfCubes &img)
// Read the image of a multivalued cubical map.
{
	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		cubicalmap Fdummy;
		return readimage (in, img, Fdummy);
	}

	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			Cube dummy;
			in >> dummy;
			in >> dummy;
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			in. get ();
			Cube q1, q2;
			in >> q1 >> q2;
			if (!in)
				throw "Can't read the file.";
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			int dim = q1. dim ();
			rectangle r (q1. coord (), q2. coord (), dim);
			const coordinate *c;
			while ((c = r. get ()) != NULL)
				img. add (Cube (c, dim));
		}
		else
			ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readimage */

/// Reads the image of a set under a multivalued cubical map.
std::istream &readimage (std::istream &in, const SetOfCubes &dom,
	SetOfCubes &img)
{
	ignorecomments (in);
	// the general format: [x,y,z] -> {[a,b,c], [d,e,f]}
	if ((in. peek () != 'S') && (in. peek () != 's'))
	{
		while (in. peek () != EOF)
		{
			if ((closingparenthesis (in. peek ()) == EOF) &&
				!std::isdigit (in. peek ()))
			{
				ignoreline (in);
				ignorecomments (in);
				continue;
			}
			Cube q;
			in >> q;
			bool ignore = !dom. check (q);
			ignorecomments (in);
			while (in. peek () == '-')
				in. get ();
			in. get (); // '>'
			ignorecomments (in);
			int opening = in. get ();
			int closing = closingparenthesis (opening);
			if (closing == EOF)
				throw "An opening brace '{' expected.";
			while ((in. peek () != EOF) &&
				(in. peek () != closing))
			{
				if (ignore)
				{
					if (in. get () == opening)
					{
						while ((in. peek () != EOF) &&
							(in. peek () != closing))
						{
							in. get ();
						}
						in. get (); // '}'
					}
				}
				else
				{
					in >> q;
					img. add (q);
					ignorecomments (in);
					if (in. peek () == ',')
					{
						in. get ();
						ignorecomments (in);
					}
				}
			}
		//	in. get (); // '}'
			ignoreline (in);
			ignorecomments (in);
		}
	}

	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			Cube domcube1, domcube2;
			in >> domcube1;
			if (!dom. check (domcube1))
			{
				ignoreline (in);
				ignorecomments (in);
				continue;
			}
			in >> domcube2;
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			in. get ();
			Cube q1, q2;
			in >> q1 >> q2;
			if (!in)
				throw "Can't read the file.";
			ignorecomments (in);
			in. get ();
			ignorecomments (in);
			if (dom. check (domcube1))
			{
				int dim = q1. dim ();
				rectangle r (q1. coord (), q2. coord (), dim);
				const coordinate *c;
				while ((c = r. get ()) != NULL)
					img. add (Cube (c, dim));
			}
		}
		else
			ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readimage */

std::istream &readselective (std::istream &in, const SetOfCubes &dom, cubicalmap &m)
{
	SetOfCubes empty;
	return readselective (in, dom, empty, m);
} /* readselective */

std::istream &readselective (std::istream &in, const SetOfCubes &dom1, const SetOfCubes &dom2,
	cubicalmap &m)
// Read a restriction of a multivalued cubical map to the given set.
// The order of arguments reversed to distinguish form the template.
{
	if (!(int) dom1 && !(int) dom2)
	{
		sout << "Warning: The domain of the map is empty.\n";
		return in;
	}

	ignorecomments (in);
	if ((in. peek () != 'S') && (in. peek () != 's'))
		return readselective (in, m, dom1, dom2);

	while (in. peek () != EOF)
	{
		if (in. peek () == '[')
		{
			in. get ();
			Cube q;
			in >> q;
			int dim = q. dim ();
			Cube domcube (q. coord (), dim);
			if (dom1. check (domcube) || dom2. check (domcube))
			{
				in >> q;
				ignorecomments (in);
				in. get ();
				ignorecomments (in);
				in. get ();
				Cube q1, q2;
				in >> q1 >> q2;
				if (!in)
					throw "Can't read the file.";
				SetOfCubes &img = m [domcube];
				rectangle r (q1. coord (), q2. coord (),
					dim);
				const coordinate *c;
				while ((c = r. get ()) != NULL)
					img. add (Cube (c, dim));
			}
		}
		ignoreline (in);
		ignorecomments (in);
	}

	return in;
} /* readselective */


// --------------------------------------------------
// ------------------ reduce cubes ------------------
// --------------------------------------------------

int neighbor2bit (const Cube &q, const Cube &neighbor)
{
	const coordinate *c0 = q, *c1 = neighbor;
	int dim = q. dim ();

	int number = 0;
	const coordinate *wrap = PointBase::wrap (dim);
	for (int i = 0; i < dim; i ++)
	{
		if (i)
			number *= 3;
		switch (c1 [i] - c0 [i])
		{
			case -1:
				number ++;
			case 1:
				number ++;
				break;
			case 0:
				break;
			default:
				if (!wrap || !wrap [i])
					return -1;
				if ((c1 [i] - c0 [i]) == (wrap [i] - 1))
					number += 2;
				else if ((c1 [i] - c0 [i]) == (1 - wrap [i]))
					number ++;
				else
					return -1;
				break;
		}
	}

	return number - 1;
} /* neighbor2bit */

Cube bit2neighbor (const Cube &q, int number, bool unconditional)
{
	coordinate c1 [Cube::MaxDim];
	const coordinate *c0 = q;
	int dim = q. dim ();

	number ++;
	for (int i = dim - 1; i >= 0; i --)
	{
		switch (number % 3)
		{
			case 2:
				c1 [i] = c0 [i];
				c1 [i] --;
				break;
			case 1:
				c1 [i] = c0 [i];
				c1 [i] ++;
				break;
			case 0:
				c1 [i] = c0 [i];
				break;
			default:
				throw "Erratic neighbor.";
		}
		number /= 3;
	}

	if (unconditional || PointBase::check (c1, dim))
		return Cube (c1, dim);
	else
		return q;
} /* bit2neighbor */

int getneighbors (const Cube &q, BitField *bits, const SetOfCubes &set,
	int limit)
{
	// if the answer is trivial, return it
	if (!(int) set)
		return 0;

	// extract the dimension of the cube and an upper bound
	// for the number of its neighbors
	int dim = q. dim ();
	int maxneighbors = getmaxneighbors (dim);

	// would it be faster to go through the entire set?
	int usetheset = (maxneighbors > (int) set);

	// prepare a counter for the number of neighbors
	int count = 0;

	// if the set is small
	if (usetheset)
	{
		// go through all the elements in the set
		for (int i = 0; i < (int) set; i ++)
		{
			// if this is the current cube, ignore it
			if (set [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}
	else
	{
		// go through all possible neighbor numbers
		for (int number = 0; number < maxneighbors; number ++)
		{
			// if it is unknown whether the neighbor exists...
			if (!bits || !bits -> test (number))
			{
				// create a neighbor cube
				Cube neighbor = bit2neighbor (q, number);
				// if not existent, ignore it
				if (neighbor == q)
					continue;
				// if this cube is not in the set, ignore it
				if (!set. check (neighbor))
					continue;
				// set the appropriate bit in the bit field
				if (bits)
					bits -> set (number);
			}
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}

	return count;
} /* getneighbors */

int getneighbors (const Cube &q, BitField *bits, const SetOfCubes &set1,
	const SetOfCubes &set2, int limit)
{
	// if the answer is trivial, return it
	if (!(int) set1 && !(int) set2)
		return 0;

	// extract the dimension of the cube and an upper bound
	// for the number of its neighbors
	int dim = q. dim ();
	int maxneighbors = getmaxneighbors (dim);

	// would it be faster to go through the entire set?
	int usetheset = (maxneighbors > (int) set1 + (int) set2);

	// prepare a counter for the number of neighbors
	int count = 0;

	// if the set is small
	if (usetheset)
	{
		// go through all the elements in the set
		for (int i = 0; i < (int) set1; i ++)
		{
			// if this is the current cube, ignore it
			if (set1 [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set1 [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
		// go through all the elements in the set
		for (int i = 0; i < (int) set2; i ++)
		{
			// if this is the current cube, ignore it
			if (set2 [i] == q)
				continue;
			// determine the number of this neighbor
			int number = neighbor2bit (q, set2 [i]);
			// if not neighbor, ignore it
			if (number < 0)
				continue;
			// set the corresponding bit in the bit field
			if (bits)
				bits -> set (number);
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}
	else
	{
		// go through all possible neighbor numbers
		for (int number = 0; number < maxneighbors; number ++)
		{
			// if it is unknown whether the neighbor exists...
			if (!bits || !bits -> test (number))
			{
				// create a neighbor cube
				Cube neighbor = bit2neighbor (q, number);
				// if not existent, ignore it
				if (neighbor == q)
					continue;
				// if this cube is not in the set, ignore it
				if (!set1. check (neighbor) &&
					!set2. check (neighbor))
					continue;
				// set the appropriate bit in the bit field
				if (bits)
					bits -> set (number);
			}
			// increase the counter
			count ++;
			// if this is enough, finish
			if (limit && (count >= limit))
				return count;
		}
	}

	return count;
} /* getneighbors */

int addneighbors (const Cube &q, const BitField &bits, SetOfCubes &set,
	const SetOfCubes &notthese)
{
	int maxneighbors = getmaxneighbors (q. dim ());

	int count = 0;
	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		Cube neighbor = bit2neighbor (q, number);
		if (!notthese. check (neighbor))
			set. add (neighbor);
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */

int addneighbors (const Cube &q, const BitField &bits, SetOfCubes &set,
	bool unconditional)
{
	int maxneighbors = getmaxneighbors (q. dim ());
	int count = 0;

	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		Cube neighbor = bit2neighbor (q, number, unconditional);
		set. add (neighbor);
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */

int addneighbors (const Cube &q, const BitField &bits,
	gcomplex<CubicalCell,integer> &c, bool unconditional)
{
	int maxneighbors = getmaxneighbors (q. dim());

	int count = 0;
	int number = bits. find (0, maxneighbors);
	while (number >= 0)
	{
		Cube neighbor = bit2neighbor (q, number, unconditional);
		c. add (CubicalCell (q, neighbor));
		number = bits. find (number + 1, maxneighbors);
		count ++;
	}

	return count;
} /* addneighbors */

inline Cube zerocube (int dim)
{
	coordinate c [Cube::MaxDim];
	for (int i = 0; i < dim; ++ i)
		c [i] = 0;
	Cube q (c, dim);
	return q;
} /* zerocube */

int acyclic (int dim, BitField &b)
{
	// look for the answer in the tabulated data
	const char *table = tabulated [dim];
	if (table)
		return Tabulated::get (table,
			bits2int (b, getmaxneighbors (dim)));

	// look for the answer among the known combinations
	SetOfBitFields *known = knownbits [dim];
	int answer = known ? known -> check (b) : -1;

	// if the answer is known, return it
	if (answer >= 0)
		return answer;

	// create a model cube for building the neighborhood
	Cube qzero = zerocube (dim);

	// find out whether the set of neighbors is acyclic or not
/*	SetOfCubes neighbors;
	addneighbors (qzero, b, neighbors);
	SetOfCubes empty;
	cubreducequiet (neighbors, empty, empty, known);
	answer = ((int) neighbors == 1);
*/
// /*
	gcomplex<CubicalCell,integer> neighbors;
	addneighbors (qzero, b, neighbors, true);
	answer = (int) acyclic (neighbors);
//*/
	// add the answer to the list of known ones
	if (known)
		known -> add (b, answer);

	return answer;
} /* acyclic */

bool acyclic (const SetOfCubes &cset)
{
	// an empty set is not acyclic
	if (!(int) cset)
		return false;

	// reduce the set and see if there is only one cube remaining
	SetOfCubes emptycubes;
	SetOfCubes theset = cset;
	cubreducequiet (theset, emptycubes, emptycubes);
	if ((int) theset == 1)
		return true;

	// create a cubical complex from this set of cubes
	cubicalcomplex ccompl;
	for (int i = 0; i < (int) cset; i ++)
		ccompl. add (CubicalCell (cset [i]));

	// check whether this geometric complex is acyclic or not
	return acyclic (ccompl);
	
/*	// OLD VERSION FOLLOWS (seems to be a repetition of what I already
	// programmed in "gcomplex" for more general data structures)

// Acknowledgements: There was a memory leak here - "hom" was not deleted.
// Thanks to Marc Niethammer for pointing this out! July 29, 2004.

	// create a cubical complex from this set of cubes and collapse it
	cubicalcomplex ccompl;
	for (int i = 0; i < (int) cset; i ++)
		ccompl. add (CubicalCell (cset [i]));
	cubicalcomplex emptycompl;
	ccompl. collapse (emptycompl, emptycompl, true, false, false, NULL,
		true);
	if ((int) ccompl == 1)
		return true;

	// create the corresponding chain complex and compute its homology
	chaincomplex<integer> cx (ccompl. dim ());
	createchaincomplex (cx, ccompl, true);
	chain<integer> *hom;
	cx. simple_form ((int *) NULL, true);
	cx. simple_homology (hom, NULL);

	// if there is anything above the zero level, the set is not acyclic
	for (int j = 1; j <= cx. dim (); j ++)
		if (hom [j]. length ())
		{
			delete [] hom;
			return false;
		}

	// if there is more than one connected component, it is not acyclic
	if (hom [0]. length () != 1)
	{
		delete [] hom;
		return false;
	}

	// if the coefficient is not 1, the set is not acyclic
	if (hom [0]. getcoefficient (0) == 1)
	{
		delete [] hom;
		return true;
	}
	else
	{
		delete [] hom;
		return false;
	}
*/
} /* acyclic */

// --------------------------------------------------

static int computeimage (SetOfCubes &img, const CubicalCell &face,
	const cubicalmap &map, const SetOfCubes &cset, const Cube &ignore)
// Compute the image of the face under the combinatorial cubical
// multivalued map, but don't take the cube 'q' into consideration.
// Return the number of cubes whose images were added to 'img'.
{
	// get the coordinates of the cubical cell
	const coordinate *left = face. leftcoord ();
	coordinate right [CubicalCell::MaxDim];
	face. rightcoord (right);

	// find the images of all the cubes containing this cell
	int spacedim = face. spacedim ();
	coordinate leftb [CubicalCell::MaxDim];
	coordinate rightb [CubicalCell::MaxDim];
	for (int j = 0; j < spacedim; j ++)
	{
		int diff = (left [j] == right [j]);
		leftb [j] = (coordinate) (left [j] - diff);
		rightb [j] = (coordinate) (right [j] + diff);
	}
	rectangle r (leftb, rightb, spacedim);
	const coordinate *c;
	int countimages = 0;
	while ((c = r. get ()) != NULL)
	{
		if (!PointBase::check (c, spacedim))
			continue;
		Cube q (c, spacedim);
		if (q == ignore)
			continue;
		if (!cset. check (q))
			continue;
		img. add (map (q));
		countimages ++;
	}

	return countimages;
} /* computeimage */

bool remainsacyclic (const cubicalmap &map, const Cube &q,
	const SetOfCubes &cset1, const SetOfCubes *cset2)
{
	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (q. dim ());

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// clear the neighborbits
	b. clearall (maxneighbors);

	// get the bitfield representing the set of the neighbors of the cube
	int n;
	if (cset2)
		n = getneighbors (q, &b, cset1, *cset2);
	else
		n = getneighbors (q, &b, cset1);

	// create all the faces of the cube
	cubicalcomplex faces;
	addneighbors (q, b, faces);
	faces. addboundaries ();

	// compute the new images of all the faces
	// and determine if they are acyclic
	int startdim = faces. dim ();
	if (startdim >= n - 1)
		startdim --;
	for (int d = startdim; d >= 0; d --)
	{
		for (int i = 0; i < (int) faces [d]; i ++)
		{
			// compute the image of the face in the first set
			SetOfCubes img;
			int n = computeimage (img, faces [d] [i], map,
				cset1, q);

			// add the image of the second set if applicable
			if (cset2)
				n += computeimage (img, faces [d] [i], map,
					*cset2, q);

			// if this is the image of only one cube, it is Ok
			if (n == 1)
				continue;

			// verify that if the large image (with 'q')
			// can be reduced towards the small image
			SetOfCubes imgsurplus = map (q);
			imgsurplus. remove (img);
			cubreducequiet (img, imgsurplus);
			if ((int) imgsurplus)
				return false;
		}
	}

	return true;
} /* remainsacyclic */

bool acyclic (const Cube &q, int dim, const SetOfCubes &cset,
	BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
		return bddacyclic (q, dim, cset, *b);

	// get the neighbors of the cube
	int n = getneighbors (q, b, cset);

	// if the answer is obvious from the number of neighbors, return it
	if ((n == 0) || (n == maxneighbors))
		return false;
	if (n == 1)
		return true;

	// return the information on whether this set of neighbors is acyclic
	return acyclic (dim, *b);
} /* acyclic */

bool acyclic (const Cube &q, int dim, const SetOfCubes &cset1,
	const SetOfCubes &cset2, BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
		return bddacyclic (q, dim, cset1, cset2, *b);

	// get the neighbors of the cube
	int n = getneighbors (q, b, cset1, cset2);

	// if the answer is obvious from the number of neighbors, return it
	if ((n == 0) || (n == maxneighbors))
		return false;
	if (n == 1)
		return true;

	// return the information on whether this set of neighbors is acyclic
	return acyclic (dim, *b);
} /* acyclic */

/// Verifies whether a cube from the other set can be removed.
inline bool acyclic_rel (const Cube &q, int dim, const SetOfCubes &cset,
	const SetOfCubes &other, BitField *b, int maxneighbors)
{
	// use a binary decision diagram code if possible
	if (dim <= MaxBddDim)
	{
		if (!getneighbors (q, b, cset, 1))
			return true;
		if (!bddacyclic (q, dim, other, *b))
			return false;
		return bddacyclic (q, dim, cset, other, *b);
	}

	// get the neighbors from the other set
	int nother = getneighbors (q, b, other);

	// verify if this set of neighbors is acyclic
	bool otheracyclic = (nother < maxneighbors) &&
		((nother == 1) || (nother && acyclic (dim, *b)));

	// add the neighbors from 'cset'
	int ncset = getneighbors (q, b, cset);

	// if there are no cubes in 'cset', then this cube is ok
	if (!ncset)
		return true;

	// if there are neighbors in 'cset' and the neighbors
	// from 'other' are not acyclic, skip this cube
	if (!otheracyclic)
		return false;

	// if there are neighbors from 'cset' and the neighbors
	// from both 'cset' and 'other' taken together
	// don't form an acyclic set, skip this cube
	if ((ncset + nother > 1) &&
		((ncset + nother == maxneighbors) ||
		!acyclic (dim, *b)))
		return false;
	return true;
} /* acyclic_rel */

// --------------------------------------------------

int Tabulated::compute (int dim)
{
	// if the dimension is wrong, do nothing
	if ((dim <= 0) || (dim >= maxdim))
		return -1;

	// allocate memory for the table if necessary
	if (!tables [dim])
	{
		tables [dim] = new char [size [dim]];
		deallocate [dim] = true;
	}

	// remember the memory buffer and temporarily set no tabulating
	char *thetable = tables [dim];
	tables [dim] = 0;

	// clear the entire table
	for (int i = 0; i < size [dim]; ++ i)
		thetable [i] = 0;

	// prepare a bit field
	int maxneighbors = getmaxneighbors (dim);
	int nconfigs = 1 << maxneighbors;
	BitField b;
	b. allocate (maxneighbors);

	// prepare a temporary cube
	coordinate c [Cube::MaxDim];
	for (int i = 0; i < dim; ++ i)
		c [i] = 0;
	Cube qzero (c, dim);

	// compute all the configurations
	sout << "Computing " << dim << "-dimensional configurations... ";
	for (int i = 1; i < nconfigs; ++ i)
	{
		int2bits (i, maxneighbors, b);
	//	gcomplex<CubicalCell,integer> neighbors;
	//	addneighbors (qzero, b, neighbors, true);
	//	if (acyclic (neighbors))
		SetOfCubes neighbors;
		addneighbors (qzero, b, neighbors, true);
		if (acyclic (qzero, dim, neighbors, &b, maxneighbors))
			Tabulated::set (thetable, i);
		if (!(i % 153791))
			scon << std::setw (12) << i <<
				"\b\b\b\b\b\b\b\b\b\b\b\b";
	}
	sout << "Completed successfully.\n";

	// deallocate the bit field
	b. free ();

	// set the tabulating table to the just computed one
	tables [dim] = thetable;
	return 0;
} /* Tabulated::compute */

// --------------------------------------------------

int cubexpand (SetOfCubes &cset, SetOfCubes &other, bool quiet)
{
	// if the case is trivial, return the answer
	if (!(int) cset || !(int) other)
		return 0;

	// remember the initial number of cubes in the main set
	int prev = (int) cset;

	// determine the dimension of the cubes
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	SetOfCubes queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the main set, move cubes to the other set
	// if possible, and add their neighbors to the queue
	for (int i = 0; i < (int) cset; i ++)
	{
		// show progress indicator if suitable
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (cset [i], dim, other, &b, maxneighbors))
			continue;
		
		// remove this cube from the queue
		queue. remove (cset [i]);

		// add neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (cset [i], &b, cset);
		addneighbors (cset [i], b, queue);

		// move the cube to the other set
		if (!quiet)
			sseq << '2' << cset [i] << '\n';
		other. add (cset [i]);
		cset. remove (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '*';
	count = 0;

	// take cubes from the queue and move them from 'cset' to 'other'
	while ((int) queue)
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		const Cube c = queue [0];
		queue. remove (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;

		// add the neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move this cube from 'cset' to 'other'
		if (!quiet)
			sseq << '2' << c << '\n';
		cset. remove (c);
		other. add (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'cset'
	return prev - (int) cset;
} /* cubexpand */

int cubexpand (SetOfCubes &cset, SetOfCubes &other, SetOfCubes &imgsrc, SetOfCubes &img,
	const cubicalmap &map, bool indexmap,
	bool checkacyclic, bool quiet)
{
	// if the case is trivial, return the answer
	if (!(int) cset || !(int) other)
		return 0;

	// remember the initial number of cubes in the main set
	int prev = (int) cset;

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	SetOfCubes queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the main set, move cubes to the other set
	// if possible, and add their neighbors to the queue
	for (int i = 0; i < (int) cset; i ++)
	{
		// show progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// take a cube from 'cset'
		const Cube &c = cset [i];

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;
		
		// take the image and reduce what is outside 'img'
		const SetOfCubes &cubimage = map (c);
		SetOfCubes ximage = cubimage;
		if (indexmap)
			ximage. add (c);
		ximage. remove (img);
		cubreducequiet (img, ximage);

		// if the image could not be reduced, skip the cube
		if ((int) ximage)
			continue;

		// make sure that the acyclicity of the map is not spoiled
		if (checkacyclic && !remainsacyclic (map, c, other))
			continue;

		// add the image of this cube to the image of the map
		if (!quiet && (int) cubimage)
		{
			sseq << "R\n";
			for (int j = 0; j < (int) cubimage; j ++)
				sseq << '2' << cubimage [j] << '\n';
			if (indexmap)
				sseq << '2' << c << '\n';
		}
		img. add (cubimage);
		if (indexmap)
			img. add (c);

		// remove from 'imgsrc' all the cubes added to 'img'
		imgsrc. remove (cubimage);
		if (indexmap)
			imgsrc. remove (c);

		// remove this cube from the queue
		queue. remove (c);

		// add neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move the cube to the other set
		if (!quiet)
			sseq << "L\n" << '2' << c << '\n';
		other. add (c);
		cset. remove (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '*';
	count = 0;

	// take cubes from the queue and move them from 'cset' to 'other'
	while ((int) queue)
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		Cube c = queue [0];
		queue. remove (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, other, &b, maxneighbors))
			continue;
		
		// take the image and reduce what is outside 'img'
		const SetOfCubes &cubimage = map (c);
		SetOfCubes ximage = cubimage;
		if (indexmap)
			ximage. add (c);
		ximage. remove (img);
		cubreducequiet (img, ximage);

		// if the image could not be reduced, skip the cube
		if ((int) ximage)
			continue;

		// make sure that the acyclicity of the map is not spoiled
		if (checkacyclic && !remainsacyclic (map, c, other))
			continue;

		// add the image of this cube to the image of the map
		if (!quiet && (int) cubimage)
		{
			sseq << "R\n";
			for (int i = 0; i < (int) cubimage; i ++)
				sseq << '2' << cubimage [i] << '\n';
			if (indexmap)
				sseq << '2' << c << '\n';
		}
		img. add (cubimage);
		if (indexmap)
			img. add (c);

		// remove from 'imgsrc' all the cubes added to 'img'
		imgsrc. remove (cubimage);
		if (indexmap)
			imgsrc. remove (c);

		// add the neighbors from 'cset' to the queue
		b. clearall (maxneighbors);
		getneighbors (c, &b, cset);
		addneighbors (c, b, queue);

		// move this cube from 'cset' to 'other'
		if (!quiet)
			sseq << "L\n" << '2' << c << '\n';
		cset. remove (c);
		other. add (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'cset'
	return prev - (int) cset;
} /* cubexpand */

int cubreducequiet (const SetOfCubes &maincset, SetOfCubes &cset, bool quiet)
{
	// remember the initial number of cubes in the set to be reduced
	int prev = (int) cset;

	// if the case is trivial, return the answer
	if (!prev)
		return 0;

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	SetOfCubes queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// remove cubes which can be removed
	// and add their neighbors to the queue
	for (int i = 0; i < (int) cset; i ++)
	{
		// show progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// take a cube to analyze
		const Cube &c = cset [i];

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, maincset, cset, &b, maxneighbors))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (c, &b, maincset, cset);
		
		// remove this cube from the queue
		queue. remove (c);

		// add its neighbors to the queue
		addneighbors (c, b, queue, maincset);

		// remove this cube from 'cset'
		if (!quiet)
			sseq << '0' << c << '\n';
		cset. remove (i);
		i --;
	}

	// add a temporary progress indicator and reset the counter
	if (!quiet)
		scon << "*";
	count = 0;

	// take cubes from the queue and remove them if possible
	while ((int) queue)
	{
		// update the progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue
		Cube c = queue [0];
		queue. remove (0);

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (c, dim, maincset, cset, &b, maxneighbors))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (c, &b, maincset, cset);

		// add the neighbors to the queue,
		// except for those who should be kept
		addneighbors (c, b, queue, maincset);

		// remove the cube from 'cset'
		if (!quiet)
			sseq << '0' << c << '\n';
		cset. remove (c);
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from 'added'
	return prev - (int) cset;
} /* cubreducequiet */

int cubreducequiet (SetOfCubes &cset, SetOfCubes &other, cubicalmap &map,
	const SetOfCubes &keep, bool quiet)
{
	// determine if the acyclicity of the map should be considered
	bool checkacyclic = ((int) map. getdomain () != 0);

	// remember the initial number of cubes in both sets
	int prev = (int) cset + (int) other;

	// if the case is trivial, return the answer
	if (!(int) cset)
	{
		if ((int) other)
		{
			SetOfCubes empty;
			other = empty;
		}
		return prev;
	}

	// determine the space dimension
	int dim = cset [0]. dim ();

	// compute the maximal number of neighbors of a cube
	int maxneighbors = getmaxneighbors (dim);

	// prepare a bitfield and allocate it if necessary
	static BitField b;
	static int _maxneighbors = 0;
	if (maxneighbors != _maxneighbors)
	{
		if (_maxneighbors > 0)
			b. free ();
		_maxneighbors = maxneighbors;
		b. allocate (maxneighbors);
	}

	// prepare a queue for cubes to check
	SetOfCubes queue;

	// prepare a counter for displaying the progress of computations
	int count = 0;

	// go through the other set, remove cubes which can be removed,
	// and add their neighbors to the queue
	int i;
	for (i = 0; i < (int) other; i ++)
	{
		// show the progress indicator
		count ++;
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";

		// if the cube should be kept, skip it
		if (keep. check (other [i]))
			continue;

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the cube cannot be removed, skip it
		if (!acyclic_rel (other [i], dim, cset, other, &b, maxneighbors))
			continue;

		// make sure that the acyclicity of the map on A is OK
		if (checkacyclic && !remainsacyclic (map, other [i], other))
			continue;

		// make sure that the acyclicity of the map on X is OK
		if (checkacyclic && !remainsacyclic (map, other [i],
			other, &cset))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (other [i], &b, cset, other);
		
		// remove this cube from the queue
		queue. remove (other [i]);

		// add its neighbors to the queue (except for 'keep')
		addneighbors (other [i], b, queue, keep);

		// remove this cube from 'other'
		if (!quiet)
			sseq << '0' << other [i] << '\n';
		other. remove (i);
		i --;
	}

	// show a temporary progress indicator and reset the counter
	if (!quiet)
		scon << '.';
	count = 0;

	// go through the main set, remove cubes which can be removed,
	// and add their neighbors to the queue
	for (i = 0; i < (int) cset; i ++)
	{
		// show progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// if this cube should be kept, skip it
		if (keep. check (cset [i]))
			continue;

		// clear the neighborbits
		b. clearall (maxneighbors);

		// if the neighborhood of the cube is not acyclic, skip it
		if (!acyclic (cset [i], dim, cset, other, &b, maxneighbors))
			continue;

		// make sure that the acyclicity of the map on X is OK
		if (checkacyclic && !remainsacyclic (map, cset [i],
			cset, &other))
			continue;

		// fill in the bit field if it has not been done yet
		if (dim <= MaxBddDim)
			getneighbors (cset [i], &b, cset, other);
		
		// remove this cube from the queue
		queue. remove (cset [i]);

		// add its neighbors to the queue
		addneighbors (cset [i], b, queue, keep);

		// remove this cube from 'cset'
		if (!quiet)
			sseq << '0' << cset [i] << '\n';
		cset. remove (i);
		i --;
	}

	// update the temporary progress indicator and reset the counter
	if (!quiet)
		scon << "\b*";
	count = 0;

	// take cubes from the queue and remove them if possible
	while ((int) queue)
	{
		// update the progress indicator
		if (!quiet && !(count % 373))
			scon << std::setw (10) << count <<
				"\b\b\b\b\b\b\b\b\b\b";
		count ++;

		// take a cube from the queue and remove it from the queue
		Cube c = queue [0];
		queue. remove (0);

		// clean the neighborbits
		b. clearall (maxneighbors);

		// if this cube belongs to the other set...
		if (other. check (c))
		{
			if (!acyclic_rel (c, dim, cset, other, &b, maxneighbors))
				continue;

			// make sure that the map's acyclicity on A is OK
			if (checkacyclic && !remainsacyclic (map, c, other))
				continue;

			// make sure that the map's acyclicity on X is OK
			if (checkacyclic && !remainsacyclic (map, c,
				other, &cset))
				continue;

			// fill in the bit field if it has not been done yet
			if (dim <= MaxBddDim)
				getneighbors (c, &b, cset, other);
		
			// add the neighbors of this cube to the
			// queue, but don't add neighbors from 'keep'
			addneighbors (c, b, queue, keep);

			// remove the cube from the other set
			if (!quiet)
				sseq << '0' << c << '\n';
			other. remove (c);
		}

		// otherwise, if this cube belongs to 'cset'...
		else
		{
			// if the neighborhood of the cube is not acyclic, skip it
			if (!acyclic (c, dim, cset, other, &b, maxneighbors))
				continue;

			// fill in the bit field if it has not been done yet
			if (dim <= MaxBddDim)
				getneighbors (c, &b, cset, other);

			// add the neighbors to the queue,
			// except for those who should be kept
			addneighbors (c, b, queue, keep);

			// remove the cube from 'cset'
			if (!quiet)
				sseq << '0' << c << '\n';
			cset. remove (c);
		}
	}

	// erase the temporary progress indicator
	if (!quiet)
		scon << "\b \b";

	// return the number of cubes removed from both sets
	return prev - (int) cset - (int) other;
} /* cubreducequiet */

int cubreducequiet (SetOfCubes &cset, SetOfCubes &other,
	const SetOfCubes &keep, bool quiet)
{
	cubicalmap emptymap;
	return cubreducequiet (cset, other, emptymap, keep, quiet);
} /* cubreducequiet */

} // namespace homology
} // namespace capd

// --------------------------------------------------


/// @}
