/// @addtogroup homology
/// @{

/////////////////////////////////////////////////////////////////////////////
/// @file homtools.cpp
///
/// @author Pawel Pilarczyk
/////////////////////////////////////////////////////////////////////////////

// Copyright (C) 1997-2007 by Pawel Pilarczyk.
//
// This file is part of the Homology Library.  This library is free software;
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License along
// with this software; see the file "license.txt".  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
// MA 02111-1307, USA.

// Started on April 16, 2003. Last revision: November 13, 2004.


#include "homology/config.h"
#include "homology/textfile.h"
#include "homology/homtools.h"

#include <iostream>
#include <fstream>
#include <iomanip>

namespace capd {
namespace homology {


// --------------------------------------------------
// --------------------- VERIFY ---------------------
// --------------------------------------------------

bool checkinclusion (const cubes &Xcubes, const cubes &Ycubes,
	const cubes &Bcubes, const char *Xname, const char *YBname)
{
	sout << "Verifying if " << Xname << " is contained in " <<
		YBname << "... ";
	bool failed = false;
	for (int i = 0; !failed && (i < (int) Xcubes); i ++)
		if (!Ycubes. check (Xcubes [i]) &&
			!Bcubes. check (Xcubes [i]))
			failed = true;
	if (failed)
		sout << "Failed.\nWARNING: The set " << Xname <<
			" is NOT contained in " << YBname << ".\n";
	else
		sout << "Passed.\n";
	return !failed;
} /* checkinclusion */

bool checkimagecontained (const cubicalmap &Fcubmap, const cubes &Xcubes,
	const cubes &Ycubes, const cubes &Bcubes,
	const char *Xname, const char *Yname)
{
	sout << "Verifying if the image of " << Xname <<
		" is contained in " << Yname << "... ";
	bool failed = false;
	for (int i = 0; !failed && (i < (int) Xcubes); i ++)
	{
		if (!Fcubmap. getdomain (). check (Xcubes [i]))
			continue;
		const cubes &cset = Fcubmap (Xcubes [i]);
		for (int j = 0; !failed && (j < (int) cset); j ++)
			if (!Ycubes. check (cset [j]) &&
				!Bcubes. check (cset [j]))
				failed = true;
	}
	if (failed)
		sout << "Failed.\nWARNING: The image of " << Xname <<
			" is NOT contained in " << Yname << ".\n";
	else
		sout << "Passed.\n";
	return !failed;
} /* checkimagecontained */

bool checkimagedisjoint (const cubicalmap &Fcubmap, const cubes &Acubes,
	const cubes &Ycubes, const char *Aname, const char *Yname)
{
	sout << "Verifying if the image of " << Aname <<
		" is disjoint from " << Yname << "... ";
	bool failed = false;
	for (int i = 0; !failed && (i < (int) Acubes); i ++)
	{
		if (!Fcubmap. getdomain (). check (Acubes [i]))
			continue;
		const cubes &cset = Fcubmap (Acubes [i]);
		for (int j = 0; !failed && (j < (int) cset); j ++)
			if (Ycubes. check (cset [j]))
				failed = true;
	}
	if (failed)
		sout << "Failed.\nSERIOUS WARNING: The image of " << Aname <<
			" is NOT disjoint from " << Yname << ".\n";
	else
		sout << "Passed.\n";
	return !failed;
} /* checkimagedisjoint */

bool checkacyclicmap (const mvcellmap<qcell,integer,cube> &Fcellcubmap,
	const char *Xname)
{
	sout << "Verifying if the map on " << Xname << " is acyclic... ";

	// retrieve the domain cell complex and analyze each dimension
	const gcomplex<qcell,integer> &dom = Fcellcubmap. getdomain ();
	int counter = 0;
	bool failed = false;
	for (int d = 0; !failed && (d <= dom. dim ()); d ++)
	{
		// retrieve the set of cells in this dimension and analyze it
		const hashedset<qcell> &qset = dom [d];
		for (int i = 0; !failed && (i < (int) qset); i ++)
		{
			// show progress indicator
			if (counter && !(counter % 373))
				scon << std::setw (10) << counter <<
					"\b\b\b\b\b\b\b\b\b\b";
			counter ++;

			// reduce the image of this cell
			const cubes &img = Fcellcubmap (qset [i]);
			if ((int) img == 1)
				continue;

			// if could not be reduced, compute the homology
			if (!acyclic (img))
				failed = true;
		}
	}
	if (failed)
		sout << "Failed.\n*** WARNING: The map on " << Xname <<
			" is NOT acyclic. ***\n"
			"*** The result of the computations "
			"may be totally wrong. ***\n";
	else
		sout << "Passed.         \n";
	return !failed;
} /* checkacyclicmap */


// --------------------------------------------------
// --------------------- REDUCE ---------------------
// --------------------------------------------------

void restrictAtoneighbors (cubes &Xcubes, cubes &Acubes,
	const char *Xname, const char *Aname, const cubes *keepcubes)
// Restrict the set of cubes 'Acubes' to these cubes which are neighbors
// of any of the cubes in 'Xcubes' and display appropriate messages.
{
	// if the set 'A' is empty, there is no point in doing anything
	if (!(int) Acubes)
		return;

	// display the message what is being done now
	sout << "Restricting " << Aname << " to the neighbors of " <<
		Xname << "\\" << Aname << "... ";

	// remember the previous number of cubes in 'A'
	int prev = (int) Acubes;

	// if the set 'X' is empty, the result is obvious
	if (!(int) Xcubes)
	{
		cubes empty;
		Acubes = empty;
	}

	// remove from 'A' these cubes which are not neighbors of 'X'
	sseq << "D 0\n";
	for (int i = 0; i < (int) Acubes; i ++)
	{
		if (keepcubes && keepcubes -> check (Acubes [i]))
			continue;
		if (getneighbors (Acubes [i], NULL, Xcubes, 1))
			continue;
		sseq << '0' << Acubes [i] << '\n';
		Acubes. remove (i);
		i --;
	}
	sseq << "D 100\n";

	// display the result
	sout << (prev - (int) Acubes) << " cubes removed, " <<
		(int) Acubes << " left in " << Aname << ".\n";

	return;
} /* restrictAtoneighbors */

void removeAfromX (cubes &Xcubes, const cubes &Acubes,
	const char *Xname, const char *Aname)
// Remove 'Acubes' from 'Xcubes'.
{
	if (!(int) Xcubes || !(int) Acubes)
		return;
	sout << "Computing " << Xname << "\\" << Aname << "... ";
	int prev = (int) Xcubes;
	Xcubes. remove (Acubes);
	sout << (prev - (int) Xcubes) << " cubes removed from " <<
		Xname << ", " << (int) Xcubes << " left.\n";
	return;
} /* removeAfromX */

void expandAinX (cubes &Xcubes, cubes &Acubes,
	const char *Xname, const char *Aname)
// Expand the other element of the pair into the main portion of the set.
{
	if (!(int) Xcubes || !(int) Acubes)
		return;
	sout << "Expanding " << Aname << " in " << Xname << "... ";
	int count = cubexpand (Xcubes, Acubes);
	sout << count << " cubes moved to " << Aname << ", " <<
		(int) Xcubes << " left in " << Xname << "\\" << Aname <<
		".\n";
	return;
} /* expandAinX */

void expandAinX (cubes &Xcubes, cubes &Acubes, cubes &Ycubes,
	cubes &Bcubes, const cubicalmap &Fcubmap,
	const char *Xname, const char *Aname, const char *Bname,
	bool indexmap, bool checkacyclic)
// Expand the other element of the pair into the main portion of the set.
{
	if (!(int) Xcubes || !(int) Acubes)
		return;
	sout << "Expanding " << Aname << " in " << Xname << "... ";
	int prevB = (int) Bcubes, prevY = (int) Ycubes;
	int count = cubexpand (Xcubes, Acubes, Ycubes, Bcubes,
		Fcubmap, indexmap, checkacyclic);
	sout << count << " moved to " << Aname << ", " << (int) Xcubes <<
		" left in " << Xname << "\\" << Aname << ", " <<
		((int) Bcubes - prevB) << " added to " << Bname << ".\n";
	if (prevY - (int) Ycubes != (int) Bcubes - prevB)
		sout << "WARNING: The image of " << Xname << "\\" << Aname <<
			" was not contained in Y. "
			"The result can be wrong!\n";
	return;
} /* expandAinX */

void reducepair (cubes &Xcubes, cubes &Acubes, const cubes &Xkeepcubes,
	const char *Xname, const char *Aname)
{
	if (!(int) Xcubes)
		return;
	sout << "Reducing full-dim cubes from ";
	if ((int) Acubes)
		sout << '(' << Xname << ',' << Aname << ")... ";
	else
		sout << Xname << "... ";
	int count = cubreduce (Xcubes, Acubes, Xkeepcubes);
	sout << count << " removed, " <<
		((int) Xcubes + (int) Acubes) << " left.\n";
	return;
} /* reducepair */

void reducepair (cubes &Xcubes, cubes &Acubes, cubicalmap &Fcubmap,
	const cubes &Xkeepcubes, const char *Xname, const char *Aname)
{
	if (!(int) Xcubes)
		return;
	sout << "Reducing cubes from ";
	if ((int) Acubes)
		sout << '(' << Xname << ',' << Aname << ") [acyclic]... ";
	else
		sout << Xname << " [acyclic]... ";
	int count = cubreduce (Xcubes, Acubes, Fcubmap, Xkeepcubes);
	sout << count << " removed, " <<
		((int) Xcubes + (int) Acubes) << " left.\n";
	return;
} /* reducepair */

void addmapimg (const cubicalmap &Fcubmap, const cubicalmap &FcubmapA,
	const cubes &Xcubes, const cubes &Acubes, cubes &Ykeepcubes,
	bool indexmap)
{
	if (!(int) Fcubmap. getdomain () && !(int) FcubmapA. getdomain ())
		return;
	sout << "Computing the image of the map... ";
	int prev = (int) Ykeepcubes;
	retrieveimage (Fcubmap, Ykeepcubes);
	retrieveimage (FcubmapA, Ykeepcubes);
	if (indexmap)
	{
		sout << "and of the inclusion... ";
		Ykeepcubes. add (Xcubes);
		Ykeepcubes. add (Acubes);
	}
	sout << ((int) Ykeepcubes - prev) << " cubes.\n";
	return;
} /* addmapimg */

void cubes2cells (cubes &Xcubes, cubicalcomplex &Xcompl,
	const char *Xname, bool deletecubes)
{
	// if there are no cubes to transform, do nothing
	if (!(int) Xcubes)
		return;

	// transform cubes into the cells of the same dimension
	int prev = (int) Xcompl;
	sout << "Transforming " << Xname << " into a set of cells... ";
	for (int i = 0; i < (int) Xcubes; i ++)
		Xcompl. add (qcell (Xcubes [i]));
	sout << ((int) Xcompl - prev) << " cells created.\n";

	// forget the set of cubes if requested to
	if (deletecubes)
	{
		cubes empty;
		Xcubes = empty;
	}

	return;
} /* cubes2cells */


// --------------------------------------------------
// ---------------- READ BITMAP FILE ----------------
// --------------------------------------------------

int ReadBitmapFile (const char *bmpname, SetOfCubes &cset,
	long mincolor, long maxcolor)
{
	// prepare a bitmap file with the vertical axis like in mathematics
	bmpfile bmp;
	bmp. invertedpicture ();

	// open the bitmap file
	if (bmp. open (bmpname) < 0)
		fileerror (bmpname);
	
	// scan the picture and produce the list of hypercubes
	coordinate c [2];
	for (c [1] = 0; c [1] < bmp. picture_height (); c [1] ++)
	{
		for (c [0] = 0; c [0] < bmp. picture_width (); c [0] ++)
		{
			long p = bmp. getpixel (c [0], c [1]);
			if ((p >= mincolor) && (p <= maxcolor))
				cset. add (Cube (c, 2));
		}
	}

	return 0;
} /* ReadBitmapFile */

} // namespace homology
} // namespace capd

/// @}
